import logging
import os
import pickle
import random
import sys

import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.keras.utils import to_categorical
from sklearn.metrics import confusion_matrix, mean_absolute_error, mean_squared_error
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler
from sklearn.linear_model import LinearRegression
from sklearn.utils import compute_class_weight
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint
from tensorflow.keras.models import save_model

from timesweeper import models

from timesweeper.plotting import plotting_utils as pu
from timesweeper.utils.gen_utils import read_config

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

logging.basicConfig()
logger = logging.getLogger("nets")
logger.setLevel("INFO")

seed = np.random.randint(1, 1e6)
random.seed(seed)
np.random.seed(seed)
tf.random.set_seed(seed)
logger.info(f"Random seed is: {seed}")


def scale_sel_coeffs(sel_coef_arr):
    """Scale training data, return scaled data and scaler to use for test data."""
    scaler = MinMaxScaler().fit(sel_coef_arr)

    return scaler


def get_data(input_pickle, data_type):
    """
    Loads data from pickle file and returns as list of labels and data.

    Args:
        input_pickle (str): Path to pickle created with make_training_features module.
        data_type (str): Determines either hfs or aft files to search for.

    Returns:
        list[str]: List of sweep labels for each sample
        np.arr: Array with all data stacked.
    """
    id_list = []
    rep_list = []
    data_list = []
    sel_coeffs = []
    sweep_types = []
    pikl_dict = pickle.load(open(input_pickle, "rb"))
    for sweep in pikl_dict.keys():
        sweep_types.append(sweep)
        for rep in pikl_dict[sweep].keys():
            try:
                data_list.append(np.array(pikl_dict[sweep][rep][data_type.lower()]))
            except:
                continue

            id_list.append(sweep)
            rep_list.append(rep)
            sel_coeffs.append(pikl_dict[sweep][rep]["sel_coeff"])

    return (
        id_list,
        rep_list,
        np.stack(data_list),
        sweep_types,
        np.array(sel_coeffs).reshape(-1, 1),
    )


def split_partitions(data, labs, sel_coeffs, reps):
    """
        Splits all data and labels into partitions for train/val/testing.

    Args:
        data (np.arr): Data for training model.
        labs (List): List of numeric labels for IDs
        sel_coeffs (List[float]): List of selection coefficients for each rep

    Returns:
        Tuple[List[narr], List[narr], List[narr], List[int], List[int], List[int]]: Train/val/test splits of IDs and labs
    """
    (
        train_data,
        val_data,
        train_labs,
        val_labs,
        train_s,
        val_s,
        train_reps,
        val_reps,
    ) = train_test_split(data, labs, sel_coeffs, reps, stratify=labs, test_size=0.3)
    (
        val_data,
        test_data,
        val_labs,
        test_labs,
        val_s,
        test_s,
        val_reps,
        test_reps,
    ) = train_test_split(
        val_data, val_labs, val_s, val_reps, stratify=val_labs, test_size=0.5
    )
    return (
        train_data,
        val_data,
        test_data,
        train_labs,
        val_labs,
        test_labs,
        train_s,
        val_s,
        test_s,
        train_reps,
        val_reps,
        test_reps,
    )


def fit_class_model(
    out_dir,
    model,
    data_type,
    train_data,
    train_labs,
    val_data,
    val_labs,
    experiment_name,
):
    """
    Fits a given model using training/validation data, plots history after done.
    Args:
        out_dir (str): Base directory where data is located, model will be saved here.
        model (Model): Compiled Keras model.
        data_type (str): Whether data is HFS or aft data.
        train_data (np.arr): Training data.
        train_labs (list[int]): OHE labels for training set.
        val_data (np.arr): Validation data.
        val_labs (list[int]): OHE labels for validation set.
        experiment_name (str): Descriptor of the sampling strategy used to generate the data. Used to ID the output.
    Returns:
        Model: Fitted Keras model, ready to be used for accuracy characterization.
    """
    monitor = "val_accuracy"
    train_target = train_labs
    val_target = val_labs

    if not os.path.exists(os.path.join(out_dir, "images")):
        os.makedirs(os.path.join(out_dir, "images"), exist_ok=True)

    if not os.path.exists(os.path.join(out_dir, "trained_models")):
        os.makedirs(os.path.join(out_dir, "trained_models"), exist_ok=True)

    checkpoint = ModelCheckpoint(
        os.path.join(out_dir, "trained_models", f"{model.name}_{data_type}"),
        monitor=monitor,
        verbose=1,
        save_best_only=True,
        save_weights_only=True,
        mode="auto",
    )

    earlystop = EarlyStopping(
        monitor=monitor,
        min_delta=0.1,
        patience=20,
        verbose=1,
        mode="auto",
        restore_best_weights=True,
    )

    callbacks_list = [earlystop, checkpoint]

    history = model.fit(
        x=train_data,
        y=train_target,
        epochs=100,
        verbose=2,
        callbacks=callbacks_list,
        validation_data=(val_data, val_target),
        # class_weight=class_weights,
    )

    pu.plot_class_training(
        os.path.join(out_dir, "images"),
        history,
        f"{experiment_name}_{model.name}_{data_type}",
    )

    # Won't checkpoint handle this?
    save_model(
        model,
        os.path.join(
            out_dir, "trained_models", f"{experiment_name}_{model.name}_{data_type}"
        ),
    )

    return model


def fit_reg_model(
    out_dir, model, data_type, train_data, train_s, val_data, val_s, experiment_name,
):
    """
    Fits a given model using training/validation data, plots history after done.

    Args:
        out_dir (str): Base directory where data is located, model will be saved here.
        model (Model): Compiled Keras model.
        data_type (str): Whether data is HFS or aft data.
        train_data (np.arr): Training data.
        train_labs (list[int]): OHE labels for training set.
        val_data (np.arr): Validation data.
        val_labs (list[int]): OHE labels for validation set.
        experiment_name (str): Descriptor of the sampling strategy used to generate the data. Used to ID the output.

    Returns:
        Model: Fitted Keras model, ready to be used for accuracy characterization.
    """
    monitor = "val_mse"
    train_target = train_s
    val_target = val_s

    if not os.path.exists(os.path.join(out_dir, "images")):
        os.makedirs(os.path.join(out_dir, "images"), exist_ok=True)

    if not os.path.exists(os.path.join(out_dir, "trained_models")):
        os.makedirs(os.path.join(out_dir, "trained_models"), exist_ok=True)

    checkpoint = ModelCheckpoint(
        os.path.join(out_dir, "trained_models", f"{model.name}_{data_type}"),
        monitor=monitor,
        verbose=1,
        save_best_only=True,
        save_weights_only=True,
        mode="auto",
    )

    earlystop = EarlyStopping(
        monitor=monitor,
        min_delta=0.01,
        patience=40,
        verbose=1,
        mode="auto",
        restore_best_weights=True,
    )

    callbacks_list = [earlystop, checkpoint]

    history = model.fit(
        x=train_data,
        y=train_target,
        epochs=100,
        verbose=2,
        callbacks=callbacks_list,
        validation_data=(val_data, val_target),
        # class_weight=class_weights,
    )

    pu.plot_reg_training(
        os.path.join(out_dir, "images"),
        history,
        f"{experiment_name}_{model.name}_{data_type}",
    )

    # Won't checkpoint handle this?
    save_model(
        model,
        os.path.join(
            out_dir, "trained_models", f"REG_{experiment_name}_{model.name}_{data_type}"
        ),
    )

    return model


def evaluate_reg_model(
    model,
    test_data,
    test_labs,
    trans_test_s,
    test_reps,
    s_scaler,
    out_dir,
    scenarios,
    experiment_name,
    data_type,
    lab_dict,
):
    """
    Evaluates model using confusion matrices and plots results.

    Args:
        model (Model): Fit Keras model.
        test_data (List[narr]): Testing data.
        test_labs (narr): Testing labels.
        test_s (narr): Selection coefficients to test against.
        out_dir (str): Base directory data is located in.
        scenarios (list[str]): Scenarios defined in config.
        experiment_name (str): Descriptor of the sampling strategy used to generate the data. Used to ID the output.
        data_type (str): Whether data is aft or hfs.
    """
    str_lab_dict = {value: key for key, value in lab_dict.items()}
    trans_pred_s = model.predict(test_data)
    trues = np.argmax(test_labs, axis=1)

    if "log" in experiment_name:
        pred_s = 10 ** (-trans_pred_s)
        test_s = 10 ** (-trans_test_s)
    elif "minmax" in experiment_name:
        pred_s = s_scaler.inverse_transform(trans_pred_s.reshape(-1, 1))
        test_s = s_scaler.inverse_transform(trans_test_s.reshape(-1, 1))
    else:
        pred_s = trans_pred_s
        test_s = trans_test_s

    pred_dict = {
        "rep": test_reps,
        "class": [str_lab_dict[i] for i in trues],
        "true_sel_coeff": test_s.flatten(),
        "pred_sel_coeff": pred_s.flatten(),
    }

    pred_df = pd.DataFrame(pred_dict)

    os.makedirs(os.path.join(out_dir, "test_predictions"), exist_ok=True)
    pred_df.to_csv(
        os.path.join(
            out_dir,
            "test_predictions",
            f"{experiment_name}_{model.name}_{data_type}_selcoeff_test_predictions.csv",
        ),
        header=True,
        index=False,
    )

    logger.info(
        f"\nMean absolute error for Sel Coeff predictions: {mean_absolute_error(test_s, pred_s)}"
    )
    pu.plot_sel_coeff_preds(
        trues,
        test_s,
        pred_s,
        os.path.join(
            out_dir,
            "images",
            f"{experiment_name}_{model.name}_{data_type}_selcoeffs.pdf",
        ),
        scenarios,
    )

    """
    # Linear regression to correct for prediction boundary
    (
        train_pred_s,
        test_pred_s,
        train_true_s,
        test_true_s,
        train_corr_rep,
        test_corr_rep,
        train_corr_class,
        test_corr_class,
    ) = train_test_split(trans_pred_s, trans_test_s, test_reps, trues, test_size=0.3)

    linreg = LinearRegression()
    linreg.fit(train_pred_s, train_true_s)
    logger.info(
        f"\nScore for s correction linear model: {linreg.score(train_pred_s, train_true_s)}"
    )

    corrected_preds = linreg.predict(test_pred_s)
    logger.info(
        f"\nMSE for  s correction linear model: {mean_squared_error(test_true_s, corrected_preds)}"
    )

    if "log" in experiment_name:
        pred_s = 10 ** (-corrected_preds)
        test_s = 10 ** (-test_true_s)
    elif "minmax" in experiment_name:
        pred_s = s_scaler.inverse_transform(corrected_preds.reshape(-1, 1))
        test_s = s_scaler.inverse_transform(test_true_s.reshape(-1, 1))
    else:
        pred_s = corrected_preds
        test_s = test_true_s

    pred_dict = {
        "rep": test_corr_rep,
        "class": [str_lab_dict[i] for i in test_corr_class],
        "true_sel_coeff": test_s.flatten(),
        "corrected_pred_sel_coeff": pred_s.flatten(),
    }
    
    pred_df = pd.DataFrame(pred_dict)

    pred_df.to_csv(
        os.path.join(
            out_dir,
            "test_predictions",
            f"{experiment_name}_{model.name}_{data_type}_corrected_selcoeff_test_predictions.csv",
        ),
        header=True,
        index=False,
    )

    logger.info(
        f"\nMean absolute error for Sel Coeff predictions: {mean_absolute_error(test_s, pred_s)}"
    )
    pu.plot_sel_coeff_preds(
        test_corr_class,
        test_s,
        pred_s,
        os.path.join(
            out_dir,
            "images",
            f"{experiment_name}_{model.name}_{data_type}_corrected_selcoeffs.pdf",
        ),
        scenarios,
    )
    """


def evaluate_class_model(
    model,
    test_data,
    test_labs,
    test_reps,
    out_dir,
    scenarios,
    experiment_name,
    data_type,
    lab_dict,
):
    """
    Evaluates model using confusion matrices and plots results.

    Args:
        model (Model): Fit Keras model.
        test_data (List[narr]): Testing data.
        test_labs (narr): Testing labels.
        test_s (narr): Selection coefficients to test against.
        out_dir (str): Base directory data is located in.
        scenarios (list[str]): Scenarios defined in config.
        experiment_name (str): Descriptor of the sampling strategy used to generate the data. Used to ID the output.
        data_type (str): Whether data is aft or hfs.
    """
    str_lab_dict = {value: key for key, value in lab_dict.items()}

    class_probs = model.predict(test_data)
    class_predictions = np.argmax(class_probs, axis=1)
    trues = np.argmax(test_labs, axis=1)

    # Cannot for the life of me figure out why memory is shared b/t functions and this
    # So it gets casted twice to break that chain
    roc_trues = np.array(list(trues))
    pr_trues = np.array(list(trues))

    pred_dict = {
        "rep": test_reps,
        "true": [str_lab_dict[i] for i in trues],
        "pred": [str_lab_dict[i] for i in class_predictions],
    }
    for str_lab in lab_dict:
        pred_dict[f"{str_lab}_prob"] = class_probs[:, lab_dict[str_lab]]

    pred_df = pd.DataFrame(pred_dict)

    os.makedirs(os.path.join(out_dir, "test_predictions"), exist_ok=True)
    pred_df.to_csv(
        os.path.join(
            out_dir,
            "test_predictions",
            f"{experiment_name}_{model.name}_{data_type}_class_test_predictions.csv",
        ),
        header=True,
        index=False,
    )

    lablist = [i.upper() for i in lab_dict]

    conf_mat = confusion_matrix(trues, class_predictions)

    pu.plot_confusion_matrix(
        os.path.join(out_dir, "images"),
        conf_mat,
        lablist,
        title=f"{experiment_name}_{model.name}_{data_type}_confmat_normed",
        normalize=True,
    )
    pu.plot_confusion_matrix(
        os.path.join(out_dir, "images"),
        conf_mat,
        lablist,
        title=f"{experiment_name}_{model.name}_{data_type}_confmat_unnormed",
        normalize=False,
    )

    pu.print_classification_report(trues, class_predictions)

    pu.plot_roc(
        roc_trues,
        class_probs,
        f"{experiment_name}_{model.name}_{data_type}",
        scenarios,
        os.path.join(
            out_dir, "images", f"{experiment_name}_{model.name}_{data_type}_roc.pdf"
        ),
    )

    pu.plot_prec_recall(
        pr_trues,
        class_probs,
        f"{experiment_name}_{model.name}_{data_type}",
        scenarios,
        os.path.join(
            out_dir, "images", f"{experiment_name}_{model.name}_{data_type}_pr.pdf"
        ),
    )


def main(ua):
    yaml_data = read_config(ua.yaml_file)
    work_dir = yaml_data["work dir"]
    os.makedirs(os.path.join(work_dir, "images"), exist_ok=True)

    # Collect all the data
    logger.info("Starting training process.")

    ids, raw_reps, raw_ts_data, sweep_types, raw_sel_coeffs = get_data(
        ua.training_data, ua.data_type
    )
    lab_dict = {str_id: int_id for int_id, str_id in enumerate(sweep_types)}

    # Convert to numerical ohe IDs
    num_ids = np.array([lab_dict[lab] for lab in ids])
    raw_ohe_ids = to_categorical(
        np.array([lab_dict[lab] for lab in ids]), len(set(ids))
    )

    if ua.subsample_amount:
        subsample_amount = ua.subsample_amount * len(set(ids))
        # Subsample to test for training size effects
        ts_data, _, ohe_ids, _, sel_coeffs, _, reps, _ = train_test_split(
            raw_ts_data,
            raw_ohe_ids,
            raw_sel_coeffs,
            raw_reps,
            train_size=subsample_amount,
            stratify=raw_ohe_ids,
        )
    else:
        ts_data = raw_ts_data
        ohe_ids = raw_ohe_ids
        reps = raw_reps
        sel_coeffs = raw_sel_coeffs

    logger.info(f"Data is subsampled to {len(ts_data)}")

    class_weights = dict(
        enumerate(
            compute_class_weight(
                class_weight="balanced", classes=np.unique(num_ids), y=num_ids
            )
        )
    )
    logger.info(f"Class weights: {class_weights}")

    if ua.data_type == "aft":
        # Needs to be in correct dims order for Conv1D layer
        datadim = ts_data.shape[1:]
        logger.info(
            f"{ua.data_type.upper()} TS Data shape (samples, timepoints, alleles): {ts_data.shape}"
        )
    else:
        logger.info(f"TS Data shape (samples, timepoints, haps): {ts_data.shape}")
        datadim = ts_data.shape[1:]
        logger.info(f"{len(ts_data)} samples in dataset.")

    logger.info("Splitting Partitions for Classification Task")
    (
        ts_train_data,
        ts_val_data,
        ts_test_data,
        train_labs,
        val_labs,
        test_labs,
        train_s,
        val_s,
        test_s,
        _train_reps,
        _val_reps,
        test_reps,
    ) = split_partitions(ts_data, ohe_ids, sel_coeffs, reps)

    # Time-series model training and evaluation
    logger.info("Training time-series model.")

    # Lazy switch for testing
    model_type = "1dcnn"
    if model_type == "1dcnn":
        class_model = models.create_TS_class_model(datadim, len(lab_dict))  # type: ignore
        reg_model = models.create_TS_reg_model(datadim)  # type: ignore
    elif model_type == "2dcnn":
        ts_train_data = np.expand_dims(ts_train_data, -1)
        ts_val_data = np.expand_dims(ts_val_data, -1)
        ts_test_data = np.expand_dims(ts_test_data, -1)
        datadim = ts_train_data.shape[1:]
        class_model = models.create_2D_TS_class_model(datadim, len(lab_dict))  # type: ignore
        reg_model = models.create_2D_TS_reg_model(datadim)  # type: ignore
    elif model_type == "chonk":
        class_model = models.create_big_TS_class_model(datadim, len(lab_dict))  # type: ignore
        reg_model = models.create_big_TS_reg_model(datadim)  # type: ignore
    elif model_type == "rnn":
        class_model = models.create_rnn_class_model(datadim, len(lab_dict))  # type: ignore
        reg_model = models.create_rnn_reg_model(datadim)  # type: ignore
    elif model_type == "transformer":
        ts_train_data = np.expand_dims(ts_train_data, -1)
        ts_val_data = np.expand_dims(ts_val_data, -1)
        ts_test_data = np.expand_dims(ts_test_data, -1)
        datadim = ts_train_data.shape[:-1]

        class_model = models.create_transformer_class_model(
            input_shape=datadim,
            head_size=256,
            num_heads=4,
            ff_dim=4,
            num_transformer_blocks=4,
            mlp_units=[128],
            mlp_dropout=0.4,
            dropout=0.25,
            n_class=len(lab_dict),
        )
        reg_model = models.create_transformer_reg_model(
            input_shape=datadim,
            head_size=256,
            num_heads=4,
            ff_dim=4,
            num_transformer_blocks=4,
            mlp_units=[128],
            mlp_dropout=0.4,
            dropout=0.25,
        )
    else:
        logger.error("Need a model")
        sys.exit(1)

    run_modes = ["class", "reg"]
    if "class" in run_modes:
        logger.info(f"\nRunning classification model for {ua.data_type}")

        trained_class_model = fit_class_model(
            work_dir,
            class_model,
            ua.data_type,
            ts_train_data,
            train_labs,
            ts_val_data,
            val_labs,
            ua.experiment_name,
        )
        evaluate_class_model(
            trained_class_model,
            ts_test_data,
            test_labs,
            test_reps,
            work_dir,
            yaml_data["scenarios"],
            ua.experiment_name,
            ua.data_type,
            lab_dict,
        )

    if "reg" in run_modes:
        for idx, scenario in enumerate(yaml_data["scenarios"][1:], start=1):
            # print(reg_model.summary())
            logger.info(
                f"{ua.data_type.upper()} Regression {scenario.upper()} TS Data shape (samples, timepoints, alleles/haplotypes): {ts_train_data.shape}"
            )
            train_idxs = np.where(
                (train_s.flatten() > 0.0) & (train_labs[:, idx] == 1)
            )[0]
            val_idxs = np.where((val_s.flatten() > 0.0) & (val_labs[:, idx] == 1))[0]
            test_idxs = np.where((test_s.flatten() > 0.0) & (test_labs[:, idx] == 1))[0]

            mode = "minmax"
            if mode == "log":
                mm_scaler = None
                trvals = -np.log10(train_s[train_idxs])
                vvals = -np.log10(val_s[val_idxs])
                tevals = -np.log10(test_s[test_idxs])
            elif mode == "minmax":
                mm_scaler = scale_sel_coeffs(train_s[train_idxs])
                pickle.dump(
                    mm_scaler,
                    open(f"{work_dir}/{ua.experiment_name}_selcoeff_scaler.pkl", "wb"),
                )
                trvals = mm_scaler.transform(train_s[train_idxs])
                vvals = mm_scaler.transform(val_s[val_idxs])
                tevals = mm_scaler.transform(test_s[test_idxs])
            else:
                mm_scaler = None
                trvals = train_s[train_idxs]
                vvals = val_s[val_idxs]
                tevals = test_s[test_idxs]

            plot = False
            if plot:
                pu.plot_s_vs_freqs(
                    train_s[train_idxs],
                    ts_train_data[train_idxs, -1, int(ts_train_data.shape[-1] / 2)]
                    - ts_train_data[train_idxs, 0, int(ts_train_data.shape[-1] / 2)],
                    scenario,
                    work_dir,
                    ua.experiment_name,
                    mode,
                )

            trained_reg_model = fit_reg_model(
                work_dir,
                reg_model,
                ua.data_type,
                ts_train_data[train_idxs],
                trvals,
                ts_val_data[val_idxs],
                vvals,
                ua.experiment_name + f"_{scenario}_{mode}",
            )
            evaluate_reg_model(
                trained_reg_model,
                ts_test_data[test_idxs],
                test_labs[test_idxs],
                tevals,
                [test_reps[i] for i in list(test_idxs)],
                mm_scaler,
                work_dir,
                yaml_data["scenarios"],
                ua.experiment_name + f"_{scenario}_{mode}",
                ua.data_type,
                lab_dict,
            )



import argparse as ap
import logging
import multiprocessing as mp
import os
import re
import subprocess
import sys
from glob import glob
from itertools import cycle

import numpy as np
import pandas as pd

from timesweeper import simulate_custom as sc
from timesweeper.utils.gen_utils import read_config

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger()
logger.setLevel("INFO")


def get_slim_code(slim_file):
    with open(slim_file, "r") as infile:
        raw_lines = [line.replace("\n", "") for line in infile.readlines()]

    return raw_lines


# fmt: off
def get_slim_info(slim_lines):
    def get_ints(lines, query):
        """Splits a line by regex split and grabs all digits. Returns as list."""
        # Uses max to handle max_years_b0, but might need to check if that causes other parsing issues
        return [
            int(s)
            for s in re.split("\W+", [i for i in lines if query in i][0])
            if s.isdigit()
        ]

    Q = get_ints(slim_lines, """defineConstant("Q",""")[0]
    gen_time = get_ints(slim_lines, """defineConstant("generation_time",""")[0]
    burn_time_mult = get_ints(slim_lines, """defineConstant("burn_in",""")[0]
    max_years_b0 = max(get_ints(slim_lines, """defineConstant("_T",""")) #Number of years BP to sim after burn in
    physLen = get_ints(slim_lines, """defineConstant("chromosome_length",""")[0]

    pop_sizes_line = [i for i in slim_lines if "_N" in i][0]
    pop_sizes_start = slim_lines.index(pop_sizes_line) + 2
    # The 4,3 will need to be more flexible
    # End of definition idx, will be end of sampling ep array
    pop_sizes_end = (slim_lines[pop_sizes_start:].index('    defineConstant("num_epochs", length(_T));') - 2 + pop_sizes_start)  

    sizes = []
    for i in slim_lines[pop_sizes_start:pop_sizes_end]:
        sizes.append([int(s) for s in re.split("\W+", i) if s.isdigit()])

    first_biggest = max([i[0] for i in sizes])
    burn_in_gens = first_biggest * burn_time_mult

    return Q, gen_time, max_years_b0, round(burn_in_gens), physLen


def inject_constants(raw_lines, sweep, recombRate, selCoeff, sel_gen, end_gen, mut_rate, dumpfile):
    """Adds in sweep type, selection coefficient, and some other details."""
    raw_lines.insert(
        raw_lines.index("    initializeMutationRate(mutation_rate);"),
        f"\tdefineConstant('sweep', '{sweep}');",
    )
    raw_lines.insert(
        raw_lines.index("    initializeMutationRate(mutation_rate);"),
        f"\tdefineConstant('selCoeff', Q * {selCoeff});",
    )
    recomb_idx = raw_lines.index("    _recombination_rates = c(")
    raw_lines[recomb_idx] = f"    _recombination_rates = c({recombRate});"
    raw_lines[recomb_idx+1] = "\n"

    raw_lines.insert(
        raw_lines.index("    initializeMutationRate(mutation_rate);"),
        f"\tinitializeMutationType('m2', 0.5, 'f', selCoeff);",
    )    
    raw_lines.insert(
        raw_lines.index("    initializeMutationRate(mutation_rate);"),
        f"\tdefineConstant('dumpFile', '{dumpfile}');"
    )

    raw_lines.insert(
        raw_lines.index("""    sim.registerLateEvent(NULL, "{dbg(self.source); end();}", G_end, G_end);"""),
        f"""    sim.registerLateEvent(NULL, "{{dbg(self.source); checkOnSweep(); }}", {sel_gen}, {end_gen});"""
    )

    raw_lines[raw_lines.index('    defineConstant("mutation_rate", Q * 0);')] = f'    defineConstant("mutation_rate", Q * {mut_rate});'
    
    return raw_lines


def sanitize_slim(raw_lines):
    """Removes sections of code inserted by stdpopsim that aren't needed."""
    raw_lines.pop(raw_lines.index("    sim.treeSeqOutput(trees_file);"))
    raw_lines.pop(raw_lines.index("    initializeTreeSeq();"))
    raw_lines.pop(
        raw_lines.index("""            "inds=p"+pop+".sampleIndividuals("+n+"); " +""")
    )
    return raw_lines


def inject_sampling(raw_lines, pop, samp_counts, gens, outfile_path):
    """Injects the actual sampling block that outputs to VCF."""
    samp_eps_line = [i for i in raw_lines if "sampling_episodes" in i][0]
    # Start of sampling_episodes constant idx
    samp_eps_start = raw_lines.index(samp_eps_line)  
    # End of definition idx, will be end of sampling ep array
    samp_eps_end = (raw_lines[samp_eps_start:].index("    ), c(3, 3)));") + samp_eps_start) 

    new_lines = []
    new_lines.extend(raw_lines[samp_eps_start : samp_eps_start + 1])
    last_line_id = len(samp_counts)
    for samps, gen, idx in zip(samp_counts, gens, range(last_line_id)):
        if idx < last_line_id - 1:
            new_lines.append("\t\t" + f"c({pop[1:]}, {samps}, {gen}),")
        else:
            new_lines.append("\t\t" + f"c({pop[1:]}, {samps}, {gen})")
    new_lines.append("\t" + f"), c(3, {len(gens)})));")

    for line in raw_lines:
        if "treeSeqRememberIndividuals" in line:
            raw_lines[
                raw_lines.index(line)
            ] = f"""\t\t\t"{pop}.outputVCFSample("+n+", replace=F, filePath='{outfile_path}', append=T);}}","""

    finished_lines = raw_lines[:samp_eps_start]
    finished_lines.extend(new_lines)
    finished_lines.extend(raw_lines[samp_eps_end + 1 :])

    return finished_lines


def make_sel_blocks(sweep, sel_gen, pop, dumpFileName):
    if sweep == "ssv":
        restart_gen = sel_gen - 500
    else:
        restart_gen = sel_gen

    intro_block = f"""
    \n{restart_gen} late(){{
        // save the state of the simulation 
        print("SAVING TO " + "{dumpFileName}" + " at generation " + sim.generation);
        sim.outputFull("{dumpFileName}");

        if (sweep == "sdn")
        {{    
            // introduce the sweep mutation
            target = sample({pop}.genomes, 1);
            target.addNewDrawnMutation(m2, asInteger(chromosome_length/2));
        }}
    }}

    """

    ssv_block = f"""
    \n{sel_gen} late(){{
        if (sweep == "ssv")
        {{
            muts = sim.mutationsOfType(m1);
            if (size(muts))
            {{
                mut = NULL;
                minDist = chromosome_length+1;
                for (m in muts)
                {{
                    freq = sim.mutationFrequencies({pop}, m);
                    if (freq > 0)
                    {{
                        dist = abs(m.position-asInteger(chromosome_length/2));
                        if (dist < minDist)
                        {{
                            minDist = dist;
                            mut = m;
                        }}
                    }}
                }}

                print("Chosen mut:" + mut.id);
                mut.setMutationType(m2);
                mut.setSelectionCoeff(selCoeff);
                
                print("Chose polymorphism at position " + mut.position + " and frequency " + sim.mutationFrequencies({pop}, mut) + " to become beneficial at generation " + sim.generation);

            }}
            else
            {{
                print("Failed to switch from neutral to beneficial at gen " + sim.generation);
            }}
        }}
    }}
    """

    check_block = f"""
    \nfunction (void)checkOnSweep(void) {{
        m1.convertToSubstitution = F;
        m2.convertToSubstitution = F;

        if (sweep == "sdn" | (sweep == "ssv"))
        {{
            fixed = (sum(sim.substitutions.mutationType == m2) == 1);
            if (fixed)
            {{
                print("FIXED in pop 1 at gen " + sim.generation);
                sim.deregisterScriptBlock(self);
            }}
            else
            {{
                muts = sim.mutationsOfType(m2);
                if (size(muts) == 0)
                {{
                    print("LOST at gen " + sim.generation + " - RESTARTING");
                    // Reload
                    sim.readFromPopulationFile("{dumpFileName}");
                
                    // Start a newly seeded run
                    setSeed(rdunif(1, 0, asInteger(2^32) - 1));
                
                    if (sweep == "sdn")
                    {{
                        // re-introduce the sweep mutation
                        target = sample({pop}.genomes, 1);
                        print("RE-INTRODUCED MUTATION at gen " + sim.generation); //" with 2Ns = " + 2*subpopSize*selCoeff);
                        target.addNewDrawnMutation(m2, asInteger(chromosome_length/2));
                    }}
                }}
            }}
        }}
    }}

    """

    all_blocks = []
    for i in intro_block, ssv_block, check_block:
        all_blocks.extend(i.split("\n"))

    return all_blocks
# fmt: on


def randomize_selCoeff_loguni(lower_bound=0.005, upper_bound=0.5):
    """Draws selection coefficient from log uniform dist to vary selection strength."""
    rng = np.random.default_rng(
        np.random.seed(int.from_bytes(os.urandom(4), byteorder="little"))
    )
    log_low = np.math.log10(lower_bound)
    log_upper = np.math.log10(upper_bound)
    rand_log = rng.uniform(log_low, log_upper, 1)

    return 10 ** rand_log[0]


def randomize_selCoeff_uni(lower_bound=0.00025, upper_bound=0.25):
    """Draws selection coefficient from log uniform dist to vary selection strength."""
    rng = np.random.default_rng(
        np.random.seed(int.from_bytes(os.urandom(4), byteorder="little"))
    )

    return rng.uniform(lower_bound, upper_bound, 1)[0]


def randomize_recombRate(lower_bound=0, upper_bound=2e-8):
    """Draws selection coefficient from log normal dist to vary selection strength."""
    rng = np.random.default_rng(
        np.random.seed(int.from_bytes(os.urandom(4), byteorder="little"))
    )
    recomb_rate = rng.uniform(lower_bound, upper_bound, 1)

    return recomb_rate[0]


def randomize_selTime(sel_time, stddev):
    """Draws from uniform dist to vary the time selection is induced before sampling in gens."""
    rng = np.random.default_rng(
        np.random.seed(int.from_bytes(os.urandom(4), byteorder="little"))
    )
    return int(rng.uniform(sel_time - stddev, sel_time + stddev, 1)[0])


def write_slim(finished_lines, slim_file, rep, work_dir):
    os.makedirs(work_dir, exist_ok=True)
    filename = os.path.basename(slim_file).split(".")[0]
    new_file_name = os.path.join(work_dir, f"modded.{rep}.{filename}.slim")
    with open(new_file_name, "w") as outfile:
        for line in finished_lines:
            outfile.write(line + "\n")

    return new_file_name


def run_slim(slimfile, slim_path):
    cmd = f"{slim_path} {slimfile}"
    try:
        subprocess.check_output(cmd.split())
    except subprocess.CalledProcessError as e:
        logger.error(e.output)

    sys.stdout.flush()
    sys.stderr.flush()


def process_vcfs(input_vcf, num_tps):
    try:
        # Split into multiples after SLiM just concats to same file
        raw_lines = sc.read_multivcf(input_vcf)
        split_lines = sc.split_multivcf(raw_lines, "##fileformat=VCFv4.2")
        if len(split_lines) > 0:
            split_lines = split_lines[len(split_lines) - num_tps :]

            # Creates subdir for each rep
            vcf_dir = sc.make_vcf_dir(input_vcf)
            sc.write_vcfs(split_lines, vcf_dir)

            # Now index and merge
            [sc.index_vcf(vcf) for vcf in glob(f"{vcf_dir}/*.vcf")]
            sc.merge_vcfs(vcf_dir)

            sc.cleanup_intermed(vcf_dir)

    except Exception as e:
        print(f"[ERROR] Couldn't process {e}")
        pass


def clean_args(ua):
    yaml_data = read_config(ua.yaml_file)
    (
        slim_file,
        reps,
        pop,
        sample_sizes,
        years_sampled,
        sel_gen,
        sel_coeff_bounds,
        mut_rate,
        work_dir,
        slim_path,
    ) = (
        yaml_data["slimfile"],
        yaml_data["reps"],
        yaml_data["pop"],
        yaml_data["sample sizes"],
        yaml_data["years sampled"],
        yaml_data["selection gen"],
        yaml_data["selection coeff bounds"],
        yaml_data["mut rate"],
        yaml_data["work dir"],
        yaml_data["slim path"],
    )

    if ua.verbose:
        logger.info(f"Number of sample sizes: {len(sample_sizes)}")
        logger.info(f"Number of years to sample from: {len(years_sampled)}")

    if len(sample_sizes) != len(years_sampled):
        logger.error(
            "Number of args supplied for generations and sample sizes don't match up. Please double check your values."
        )
        raise ValueError()

    if years_sampled[0] < years_sampled[-1]:
        logger.warning(
            "Gens were not supplied in earliest to latest order, sorting and flipping."
        )
        years_sampled = sorted(years_sampled)[::-1]
        sample_sizes = sorted(sample_sizes)[::-1]
    else:
        years_sampled = years_sampled
        sample_sizes = sample_sizes

    return (
        ua.verbose,
        ua.threads,
        ua.rep_range,
        work_dir,
        slim_file,
        reps,
        pop,
        sample_sizes,
        years_sampled,
        sel_gen,
        sel_coeff_bounds,
        mut_rate,
        slim_path,
    )


def main(ua):
    (
        verbose,
        threads,
        rep_range,
        work_dir,
        slim_file,
        reps,
        pop,
        sample_sizes,
        years_sampled,
        sel_gen,
        sel_coeff_bounds,
        mut_rate,
        slim_path,
    ) = clean_args(ua)

    work_dir = work_dir
    vcf_dir = f"{work_dir}/vcfs"
    dumpfile_dir = f"{work_dir}/dumpfiles"
    script_dir = f"{work_dir}/scripts"

    sweeps = ["neut", "sdn", "ssv"]

    for i in [vcf_dir, dumpfile_dir, script_dir]:
        for sweep in sweeps:
            os.makedirs(f"{i}/{sweep}", exist_ok=True)

    # Inject info into SLiM script and then simulate, store params for reproducibility
    sim_params = []
    script_list = []

    if rep_range:  # Take priority
        replist = range(int(rep_range[0]), int(rep_range[1]))
    else:
        replist = range(reps)

    for rep in replist:
        for sweep in sweeps:

            # Info scraping and calculations
            raw_lines = get_slim_code(slim_file)
            raw_lines = sanitize_slim(raw_lines)

            Q, gen_time, max_years_b0, burn_in_gens, physLen = get_slim_info(raw_lines)

            # Pull from variable time of selection before sampling to make more robust
            rand_sel_gen = randomize_selTime(sel_gen, 200 / Q)

            burn_in_gens = int(round(burn_in_gens / Q))

            # Convert from earliest year from bp to gens
            end_gen = int(round((max_years_b0) / gen_time / Q))

            # Written out each step for clarity, sdn to keep track of otherwise
            # Find the earliest time before present, convert to useable times
            furthest_from_pres = max(years_sampled)
            abs_year_beg = max_years_b0 - furthest_from_pres

            sel_gen_time = (
                int(((abs_year_beg / gen_time) - rand_sel_gen) / Q)
            ) + burn_in_gens

            if sel_coeff_bounds[0] == sel_coeff_bounds[1]:
                sel_coeff = sel_coeff_bounds[0] * Q
            else:
                sel_coeff = randomize_selCoeff_uni(*sel_coeff_bounds) * Q

            recombRate = randomize_recombRate()

            # Logger vars
            if verbose:
                logger.info("Timesweeper SLiM Injection")
                logger.info(f"Q Scaling Value: {Q}")
                logger.info(f"Gen Time: {gen_time}")
                logger.info(f"Simulated Chrom Length: {physLen}")
                logger.info(f"Burn in years: {burn_in_gens * gen_time}")
                logger.info(f"Burn in gens: {burn_in_gens}")
                logger.info(f"Number Years Simulated (post-burn): {max_years_b0}")
                logger.info(f"Number Gens Simulated (post-burn): {end_gen}")
                logger.info(
                    f"Number Years Simulated (inc. burn): {max_years_b0 + (burn_in_gens * gen_time)}"
                )
                logger.info(
                    f"Number gens simulated (inc. burn): {end_gen + burn_in_gens}"
                )
                logger.info(f"Selection type: {sweep}")
                logger.info(f"Selection coeff: {sel_coeff}")
                logger.info(f"Recomb Rate: {recombRate}")
                logger.info(f"Selection start gen: {sel_gen_time}")
                logger.info(f"Number of timepoints: {len(years_sampled)}")
                logger.info(
                    f"""Sample sizes (individuals): 
                    {" ".join([str(i) for i in sample_sizes])}"""
                )
                logger.info(
                    f"Years before present (1950) sampled: {' '.join([str(i) for i in years_sampled])}"
                )
                logger.info(
                    f"Gens before present (1950) sampled: {' '.join([str(int(i / gen_time / Q)) for i in years_sampled])}",
                )

            sim_params.append(
                (
                    sweep,
                    rep,
                    Q,
                    gen_time,
                    physLen,
                    burn_in_gens,
                    max_years_b0 + (burn_in_gens * gen_time),
                    sel_coeff,
                    sel_gen_time,
                    years_sampled,
                    sample_sizes,
                )
            )

            dumpfile = f"{dumpfile_dir}/{sweep}/{rep}.dump"

            # Injection
            prepped_lines = inject_constants(
                raw_lines,
                sweep,
                recombRate,
                sel_coeff,
                sel_gen_time,
                end_gen + burn_in_gens,
                mut_rate,
                dumpfile,
            )

            sampling_lines = inject_sampling(
                prepped_lines,
                pop,
                sample_sizes,
                years_sampled,
                f"{vcf_dir}/{sweep}/{rep}.multivcf",
            )

            selection_lines = make_sel_blocks(sweep, sel_gen_time, pop, dumpfile)
            finished_lines = []
            finished_lines.extend(sampling_lines)
            finished_lines.extend(selection_lines)

            script_path = write_slim(
                finished_lines, slim_file, rep, f"{script_dir}/{sweep}"
            )
            script_list.append(script_path)

    print(f"Reps simulated: {replist}")

    for script in script_list:
        run_slim(script, slim_path)

    # Process VCFs and Cleanup
    for rep in replist:
        for sweep in sweeps:
            vcf_file = f"{vcf_dir}/{sweep}/{rep}.multivcf"
            process_vcfs(vcf_file, len(sample_sizes))
            os.remove(vcf_file)
            dumpFile = f"{dumpfile_dir}/{sweep}/{rep}.dump"
            os.remove(dumpFile)

    # Save params
    if not os.path.exists(f"{work_dir}/params"):
        os.makedirs(f"{work_dir}/params")
    params_df = pd.DataFrame(
        sim_params,
        columns=[
            "sweep",
            "rep",
            "Q",
            "gen_time",
            "chrom_len",
            "burn_in_gens",
            "num_gens_simmed",
            "sel_coeff",
            "sel_gen",
            "years_bp_sampled",
            "samp_sizes",
        ],
    )
    params_df.to_csv(
        f"{work_dir}/params/sim_{'_'.join(rep_range)}_params.tsv", sep="\t", index=False
    )

    logger.info(f"Simulations finished, parameters saved to {work_dir}/sim_params.csv.")



import argparse
import os

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


def main(ua):
    if not os.path.exists(ua.output_dir):
        os.makedirs(ua.output_dir)

    log_df = pd.read_csv(ua.summary_tsv, header=0, sep="\t")
    acc_df = pd.read_csv(ua.test_file, header=0, sep=",")
    merged = pd.merge(
        right=acc_df,
        left=log_df,
        how="right",
        left_on=["rep", "sweep"],
        right_on=["rep", "class"],
    )
    merged["abs_error"] = np.abs(merged["pred_sel_coeff"] - merged["true_sel_coeff"])
    merged.to_csv(
        f"{ua.output_dir}/merged_table.tsv", index=False, header=True, sep="\t"
    )

    merged.plot(
        x="selCoeff",
        y="abs_error",
        kind="scatter",
        title="Absolute S Prediction Error vs Selection Coefficient",
        xlabel="Selection Coefficient s",
        ylabel="Absolute Error | abs(predicted s - true s)",
        xlim=(0.0, 0.25),
        ylim=(0.0, 0.25),
    )
    plt.savefig(f"{ua.output_dir}/error_vs_selcoeff.png")
    plt.clf()

    merged.plot(
        x="sampOffset",
        y="abs_error",
        kind="scatter",
        title="Absolute S Prediction Error vs Selection Coefficient",
        xlabel="Sampling Offset from Selection",
        ylabel="Absolute Error | abs(predicted s - true s)",
        xlim=(-50, 50),
        ylim=(0.0, 0.25),
        xticks=range(-50, 60, 10),
    )
    plt.savefig(f"{ua.output_dir}/error_vs_sampletime.png")
    plt.clf()

    merged.plot(
        x="sampOffset",
        y="selCoeff",
        kind="scatter",
        title="Sampling Offset from Selection vs Selection Coefficient",
        xlabel="Sampling Offset from Selection",
        ylabel="Selection Coefficient",
        xlim=(-50, 50),
        ylim=(0.0, 0.25),
        xticks=range(-50, 60, 10),
    )
    plt.savefig(f"{ua.output_dir}/selcoeff_vs_sampletime.png")
    plt.clf()



import argparse as ap
import logging
import os
from itertools import cycle

import numpy as np
import pandas as pd
import yaml
from tensorflow.keras.models import load_model
from tqdm import tqdm

from utils.gen_utils import write_preds

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

logging.basicConfig()
logger = logging.getLogger("timesweeper")
logger.setLevel("INFO")

pd.options.display.float_format = "{:.2f}".format


def load_npz(npz_path):
    npz_obj = np.load(npz_path)
    aft = npz_obj["aftIn"]  # (snps, timepoints, window)
    locs = npz_obj["aftInPosition"]  # (snps, window_locs)

    return aft, locs


def parse_npz_name(npz_path):
    splitpath = npz_path.split("_")
    chrom = splitpath[2]
    rep = splitpath[-1].split(".")[0]

    return chrom, rep


def run_aft_windows(ts_aft, locs, chrom, model):
    """
    Iterates through windows of MAF time-series matrix and predicts using NN.
    Args:
        snps (list[tup(chrom, pos,  mut)]): Tuples of information for each SNP. Contains mut only if benchmarking == True.
        genos (allel.GenotypeArray): Genotypes of all samples.
        samp_sizes (list[int]): Number of chromosomes sampled at each timepoint.
        win_size (int): Number of SNPs to use for each prediction. Needs to match how NN was trained.
        model (Keras.model): Keras model to use for prediction.
    Returns:
        dict: Prediction values in the form of dict[snps[center]]
        np.arr: the central-most window, either based on mutation type or closest to half size of chrom.
    """
    left_edges = list(locs[:, 0])
    right_edges = list(locs[:, -1])
    centers = list(locs[:, 25])
    class_probs, sel_pred = model.predict(ts_aft)

    results_list = zip(
        cycle([chrom]), centers, left_edges, right_edges, class_probs, sel_pred
    )

    return results_list


def load_nn(model_path, summary=False):
    """
    Loads the trained Keras network.
    Args:
        model_path (str): Path to Keras model.
        summary (bool, optional): Whether to print out model summary or not. Defaults to False.
    Returns:
        Keras.model: Trained Keras model to use for prediction.
    """
    model = load_model(model_path)
    if summary:
        print(model.summary())

    return model


def write_preds(results_list, outfile, benchmark):
    """
    Writes NN predictions to file.
    Args:
        results_dict (dict): SNP NN prediction scores and window edges.
        outfile (str): File to write results to.
    """
    lab_dict = {0: "Neut", 1: "SSV"}
    chrom, centers, left_edges, right_edges, probs = zip(*results_list)

    neut_scores = [i[0] for i in probs]
    ssv_scores = [i[1] for i in probs]
    classes = [lab_dict[np.argmax(i)] for i in probs]

    predictions = pd.DataFrame(
        {
            "Chrom": chrom,
            "BP": centers,
            "Class": classes,
            "Neut_Score": neut_scores,
            "SSV_Score": ssv_scores,
            "Win_Start": left_edges,
            "Win_End": right_edges,
        }
    )

    # predictions = predictions[predictions["Neut_Score"] < 0.5]

    predictions.sort_values(["Chrom", "BP"], inplace=True)

    predictions.to_csv(outfile, header=True, index=False, sep="\t", float_format="%.3f")

    bed_df = predictions[["Chrom", "Win_Start", "Win_End", "BP"]]
    bed_df.to_csv(outfile.replace(".csv", ".bed"), header=False, index=False, sep="\t")


def main(ua):
    outdir, aft_model = (
        ua.outdir,
        load_nn(ua.aft_model),
    )

    try:
        if not os.path.exists(outdir):
            os.makedirs(outdir)
    except:
        # running in high-parallel sometimes it errors when trying to check/create simultaneously
        pass

    # Load in everything
    logger.info(f"Loading data from {ua.input_file}")
    chrom, rep = parse_npz_name(ua.input_file)
    ts_aft, locs = load_npz(ua.input_file)

    # aft
    logger.info("Predicting with AFT")
    aft_predictions = run_aft_windows(ts_aft, locs, chrom, aft_model)
    write_preds(aft_predictions, f"{outdir}/aft_{chrom}_{rep}_preds.csv", ua.benchmark)
    logger.info(f"Done, results written to {outdir}/aft_{chrom}_{rep}_preds.csv")



"""
This is a modified version of the `train_nets.py` module in the core Timesweeper package. 
Modifications include:
- Fully connected network instead of 1DCNN
"""

import logging
import os
import pickle
import random
import sys

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.keras.utils import to_categorical
from sklearn.metrics import confusion_matrix, mean_absolute_error, mean_squared_error
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler
from sklearn.linear_model import LinearRegression
from sklearn.utils import compute_class_weight
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint
from tensorflow.keras.models import save_model

from timesweeper import models

from timesweeper.plotting import plotting_utils as pu
from timesweeper.utils.gen_utils import read_config

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

logging.basicConfig()
logger = logging.getLogger("nets")
logger.setLevel("INFO")

seed = np.random.randint(1, 1e6)
random.seed(seed)
np.random.seed(seed)
tf.random.set_seed(seed)
logger.info(f"Random seed is: {seed}")


def scale_sel_coeffs(sel_coef_arr):
    """Scale training data, return scaled data and scaler to use for test data."""
    scaler = MinMaxScaler().fit(sel_coef_arr)

    return scaler


def get_data(input_pickle, data_type):
    """
    Loads data from pickle file and returns as list of labels and data.

    Args:
        input_pickle (str): Path to pickle created with make_training_features module.
        data_type (str): Determines either hfs or aft files to search for.

    Returns:
        list[str]: List of sweep labels for each sample
        np.arr: Array with all data stacked.
    """
    id_list = []
    rep_list = []
    data_list = []
    sel_coeffs = []
    sweep_types = []
    pikl_dict = pickle.load(open(input_pickle, "rb"))
    for sweep in pikl_dict.keys():
        sweep_types.append(sweep)
        for rep in pikl_dict[sweep].keys():
            try:
                data_list.append(np.array(pikl_dict[sweep][rep][data_type.lower()]))
            except:
                continue
            id_list.append(sweep)
            rep_list.append(rep)
            sel_coeffs.append(pikl_dict[sweep][rep]["sel_coeff"])

    return (
        id_list,
        rep_list,
        np.stack(data_list),
        sweep_types,
        np.array(sel_coeffs).reshape(-1, 1),
    )


def split_partitions(data, labs, sel_coeffs, reps):
    """
        Splits all data and labels into partitions for train/val/testing.

    Args:
        data (np.arr): Data for training model.
        labs (List): List of numeric labels for IDs
        sel_coeffs (List[float]): List of selection coefficients for each rep

    Returns:
        Tuple[List[narr], List[narr], List[narr], List[int], List[int], List[int]]: Train/val/test splits of IDs and labs
    """
    (
        train_data,
        val_data,
        train_labs,
        val_labs,
        train_s,
        val_s,
        train_reps,
        val_reps,
    ) = train_test_split(data, labs, sel_coeffs, reps, stratify=labs, test_size=0.3)
    (
        val_data,
        test_data,
        val_labs,
        test_labs,
        val_s,
        test_s,
        val_reps,
        test_reps,
    ) = train_test_split(
        val_data, val_labs, val_s, val_reps, stratify=val_labs, test_size=0.5
    )
    return (
        train_data,
        val_data,
        test_data,
        train_labs,
        val_labs,
        test_labs,
        train_s,
        val_s,
        test_s,
        train_reps,
        val_reps,
        test_reps,
    )


def fit_class_model(
    out_dir,
    model,
    data_type,
    train_data,
    train_labs,
    val_data,
    val_labs,
    experiment_name,
):
    """
    Fits a given model using training/validation data, plots history after done.
    Args:
        out_dir (str): Base directory where data is located, model will be saved here.
        model (Model): Compiled Keras model.
        data_type (str): Whether data is HFS or aft data.
        train_data (np.arr): Training data.
        train_labs (list[int]): OHE labels for training set.
        val_data (np.arr): Validation data.
        val_labs (list[int]): OHE labels for validation set.
        experiment_name (str): Descriptor of the sampling strategy used to generate the data. Used to ID the output.
    Returns:
        Model: Fitted Keras model, ready to be used for accuracy characterization.
    """
    monitor = "val_accuracy"
    train_target = train_labs
    val_target = val_labs

    if not os.path.exists(os.path.join(out_dir, "images")):
        os.makedirs(os.path.join(out_dir, "images"), exist_ok=True)

    if not os.path.exists(os.path.join(out_dir, "trained_models")):
        os.makedirs(os.path.join(out_dir, "trained_models"), exist_ok=True)

    checkpoint = ModelCheckpoint(
        os.path.join(out_dir, "trained_models", f"{model.name}_{data_type}"),
        monitor=monitor,
        verbose=1,
        save_best_only=True,
        save_weights_only=True,
        mode="auto",
    )

    earlystop = EarlyStopping(
        monitor=monitor,
        min_delta=0.1,
        patience=20,
        verbose=1,
        mode="auto",
        restore_best_weights=True,
    )

    callbacks_list = [earlystop, checkpoint]

    history = model.fit(
        x=train_data,
        y=train_target,
        epochs=100,
        verbose=2,
        callbacks=callbacks_list,
        validation_data=(val_data, val_target),
        # class_weight=class_weights,
    )

    pu.plot_class_training(
        os.path.join(out_dir, "images"),
        history,
        f"{experiment_name}_{model.name}_{data_type}",
    )

    # Won't checkpoint handle this?
    save_model(
        model,
        os.path.join(
            out_dir, "trained_models", f"{experiment_name}_{model.name}_{data_type}"
        ),
    )

    return model


def fit_reg_model(
    out_dir, model, data_type, train_data, train_s, val_data, val_s, experiment_name,
):
    """
    Fits a given model using training/validation data, plots history after done.

    Args:
        out_dir (str): Base directory where data is located, model will be saved here.
        model (Model): Compiled Keras model.
        data_type (str): Whether data is HFS or aft data.
        train_data (np.arr): Training data.
        train_labs (list[int]): OHE labels for training set.
        val_data (np.arr): Validation data.
        val_labs (list[int]): OHE labels for validation set.
        experiment_name (str): Descriptor of the sampling strategy used to generate the data. Used to ID the output.

    Returns:
        Model: Fitted Keras model, ready to be used for accuracy characterization.
    """
    monitor = "val_mse"
    train_target = train_s
    val_target = val_s

    if not os.path.exists(os.path.join(out_dir, "images")):
        os.makedirs(os.path.join(out_dir, "images"), exist_ok=True)

    if not os.path.exists(os.path.join(out_dir, "trained_models")):
        os.makedirs(os.path.join(out_dir, "trained_models"), exist_ok=True)

    checkpoint = ModelCheckpoint(
        os.path.join(out_dir, "trained_models", f"{model.name}_{data_type}"),
        monitor=monitor,
        verbose=1,
        save_best_only=True,
        save_weights_only=True,
        mode="auto",
    )

    earlystop = EarlyStopping(
        monitor=monitor,
        min_delta=0.01,
        patience=40,
        verbose=1,
        mode="auto",
        restore_best_weights=True,
    )

    callbacks_list = [earlystop, checkpoint]

    history = model.fit(
        x=train_data,
        y=train_target,
        epochs=100,
        verbose=2,
        callbacks=callbacks_list,
        validation_data=(val_data, val_target),
        # class_weight=class_weights,
    )

    pu.plot_reg_training(
        os.path.join(out_dir, "images"),
        history,
        f"{experiment_name}_{model.name}_{data_type}",
    )

    # Won't checkpoint handle this?
    save_model(
        model,
        os.path.join(
            out_dir, "trained_models", f"REG_{experiment_name}_{model.name}_{data_type}"
        ),
    )

    return model


def evaluate_reg_model(
    model,
    test_data,
    test_labs,
    trans_test_s,
    test_reps,
    s_scaler,
    out_dir,
    scenarios,
    experiment_name,
    data_type,
    lab_dict,
):
    """
    Evaluates model using confusion matrices and plots results.

    Args:
        model (Model): Fit Keras model.
        test_data (List[narr]): Testing data.
        test_labs (narr): Testing labels.
        test_s (narr): Selection coefficients to test against.
        out_dir (str): Base directory data is located in.
        scenarios (list[str]): Scenarios defined in config.
        experiment_name (str): Descriptor of the sampling strategy used to generate the data. Used to ID the output.
        data_type (str): Whether data is aft or hfs.
    """
    str_lab_dict = {value: key for key, value in lab_dict.items()}
    trans_pred_s = model.predict(test_data)
    trues = np.argmax(test_labs, axis=1)

    if "log" in experiment_name:
        pred_s = 10 ** (-trans_pred_s)
        test_s = 10 ** (-trans_test_s)
    elif "minmax" in experiment_name:
        pred_s = s_scaler.inverse_transform(trans_pred_s.reshape(-1, 1))
        test_s = s_scaler.inverse_transform(trans_test_s.reshape(-1, 1))
    else:
        pred_s = trans_pred_s
        test_s = trans_test_s

    pred_dict = {
        "rep": test_reps,
        "class": [str_lab_dict[i] for i in trues],
        "true_sel_coeff": test_s.flatten(),
        "pred_sel_coeff": pred_s.flatten(),
    }

    pred_df = pd.DataFrame(pred_dict)

    os.makedirs(os.path.join(out_dir, "test_predictions"), exist_ok=True)
    pred_df.to_csv(
        os.path.join(
            out_dir,
            "test_predictions",
            f"{experiment_name}_{model.name}_{data_type}_selcoeff_test_predictions.csv",
        ),
        header=True,
        index=False,
    )

    logger.info(
        f"\nMean absolute error for Sel Coeff predictions: {mean_absolute_error(test_s, pred_s)}"
    )
    pu.plot_sel_coeff_preds(
        trues,
        test_s,
        pred_s,
        os.path.join(
            out_dir,
            "images",
            f"{experiment_name}_{model.name}_{data_type}_selcoeffs.pdf",
        ),
        scenarios,
    )

    # Linear regression to correct for prediction boundary
    (
        train_pred_s,
        test_pred_s,
        train_true_s,
        test_true_s,
        train_corr_rep,
        test_corr_rep,
        train_corr_class,
        test_corr_class,
    ) = train_test_split(trans_pred_s, trans_test_s, test_reps, trues, test_size=0.3)

    linreg = LinearRegression()
    linreg.fit(train_pred_s, train_true_s)
    logger.info(
        f"\nScore for s correction linear model: {linreg.score(train_pred_s, train_true_s)}"
    )

    corrected_preds = linreg.predict(test_pred_s)
    logger.info(
        f"\nMSE for  s correction linear model: {mean_squared_error(test_true_s, corrected_preds)}"
    )

    if "log" in experiment_name:
        pred_s = 10 ** (-corrected_preds)
        test_s = 10 ** (-test_true_s)
    elif "minmax" in experiment_name:
        pred_s = s_scaler.inverse_transform(corrected_preds.reshape(-1, 1))
        test_s = s_scaler.inverse_transform(test_true_s.reshape(-1, 1))
    else:
        pred_s = corrected_preds
        test_s = test_true_s

    pred_dict = {
        "rep": test_corr_rep,
        "class": [str_lab_dict[i] for i in test_corr_class],
        "true_sel_coeff": test_s.flatten(),
        "corrected_pred_sel_coeff": pred_s.flatten(),
    }

    pred_df = pd.DataFrame(pred_dict)

    pred_df.to_csv(
        os.path.join(
            out_dir,
            "test_predictions",
            f"{experiment_name}_{model.name}_{data_type}_corrected_selcoeff_test_predictions.csv",
        ),
        header=True,
        index=False,
    )

    logger.info(
        f"\nMean absolute error for Sel Coeff predictions: {mean_absolute_error(test_s, pred_s)}"
    )
    pu.plot_sel_coeff_preds(
        test_corr_class,
        test_s,
        pred_s,
        os.path.join(
            out_dir,
            "images",
            f"{experiment_name}_{model.name}_{data_type}_corrected_selcoeffs.pdf",
        ),
        scenarios,
    )


def evaluate_class_model(
    model,
    test_data,
    test_labs,
    test_reps,
    out_dir,
    scenarios,
    experiment_name,
    data_type,
    lab_dict,
):
    """
    Evaluates model using confusion matrices and plots results.

    Args:
        model (Model): Fit Keras model.
        test_data (List[narr]): Testing data.
        test_labs (narr): Testing labels.
        test_s (narr): Selection coefficients to test against.
        out_dir (str): Base directory data is located in.
        scenarios (list[str]): Scenarios defined in config.
        experiment_name (str): Descriptor of the sampling strategy used to generate the data. Used to ID the output.
        data_type (str): Whether data is aft or hfs.
    """
    str_lab_dict = {value: key for key, value in lab_dict.items()}

    class_probs = model.predict(test_data)
    class_predictions = np.argmax(class_probs, axis=1)
    trues = np.argmax(test_labs, axis=1)

    # Cannot for the life of me figure out why memory is shared b/t functions and this
    # So it gets casted twice to break that chain
    roc_trues = np.array(list(trues))
    pr_trues = np.array(list(trues))

    pred_dict = {
        "rep": test_reps,
        "true": [str_lab_dict[i] for i in trues],
        "pred": [str_lab_dict[i] for i in class_predictions],
    }
    for str_lab in lab_dict:
        pred_dict[f"{str_lab}_scores"] = class_probs[:, lab_dict[str_lab]]

    pred_df = pd.DataFrame(pred_dict)

    os.makedirs(os.path.join(out_dir, "test_predictions"), exist_ok=True)
    pred_df.to_csv(
        os.path.join(
            out_dir,
            "test_predictions",
            f"{experiment_name}_{model.name}_{data_type}_class_test_predictions.csv",
        ),
        header=True,
        index=False,
    )

    lablist = [i.upper() for i in lab_dict]

    conf_mat = confusion_matrix(trues, class_predictions)

    pu.plot_confusion_matrix(
        os.path.join(out_dir, "images"),
        conf_mat,
        lablist,
        title=f"{experiment_name}_{model.name}_{data_type}_confmat_normed",
        normalize=True,
    )
    pu.plot_confusion_matrix(
        os.path.join(out_dir, "images"),
        conf_mat,
        lablist,
        title=f"{experiment_name}_{model.name}_{data_type}_confmat_unnormed",
        normalize=False,
    )

    pu.print_classification_report(trues, class_predictions)

    pu.plot_roc(
        roc_trues,
        class_probs,
        f"{experiment_name}_{model.name}_{data_type}",
        scenarios,
        os.path.join(
            out_dir, "images", f"{experiment_name}_{model.name}_{data_type}_roc.pdf"
        ),
    )

    pu.plot_prec_recall(
        pr_trues,
        class_probs,
        f"{experiment_name}_{model.name}_{data_type}",
        scenarios,
        os.path.join(
            out_dir, "images", f"{experiment_name}_{model.name}_{data_type}_pr.pdf"
        ),
    )


def main(ua):
    yaml_data = read_config(ua.yaml_file)
    work_dir = yaml_data["work dir"]
    os.makedirs(os.path.join(work_dir, "images"), exist_ok=True)

    # Collect all the data
    logger.info("Starting training process.")

    ids, raw_reps, raw_ts_data, sweep_types, raw_sel_coeffs = get_data(
        ua.training_data, ua.data_type
    )
    lab_dict = {str_id: int_id for int_id, str_id in enumerate(sweep_types)}

    # Convert to numerical ohe IDs
    num_ids = np.array([lab_dict[lab] for lab in ids])
    raw_ohe_ids = to_categorical(
        np.array([lab_dict[lab] for lab in ids]), len(set(ids))
    )

    if ua.subsample_amount:
        subsample_amount = ua.subsample_amount * len(set(ids))
        # Subsample to test for training size effects
        ts_data, _, ohe_ids, _, sel_coeffs, _, reps, _ = train_test_split(
            raw_ts_data,
            raw_ohe_ids,
            raw_sel_coeffs,
            raw_reps,
            train_size=subsample_amount,
            stratify=raw_ohe_ids,
        )
    else:
        ts_data = raw_ts_data
        ohe_ids = raw_ohe_ids
        reps = raw_reps
        sel_coeffs = raw_sel_coeffs

    ts_data = np.squeeze(ts_data)

    logger.info(f"Data is subsampled to {len(ts_data)}")

    class_weights = dict(
        enumerate(
            compute_class_weight(
                class_weight="balanced", classes=np.unique(num_ids), y=num_ids
            )
        )
    )
    logger.info(f"Class weights: {class_weights}")

    if ua.data_type == "aft":
        # Needs to be in correct dims order for Conv1D layer
        datadim = ts_data.shape[1:]
        logger.info(
            f"{ua.data_type.upper()} TS Data shape (samples, timepoints, alleles): {ts_data.shape}"
        )
    else:
        logger.info(f"TS Data shape (samples, timepoints, haps): {ts_data.shape}")
        datadim = ts_data.shape[1:]
        logger.info(f"{len(ts_data)} samples in dataset.")

    logger.info("Splitting Partitions for Classification Task")
    (
        ts_train_data,
        ts_val_data,
        ts_test_data,
        train_labs,
        val_labs,
        test_labs,
        train_s,
        val_s,
        test_s,
        _train_reps,
        _val_reps,
        test_reps,
    ) = split_partitions(ts_data, ohe_ids, sel_coeffs, reps)

    # Time-series model training and evaluation
    logger.info("Training time-series model.")

    class_model = models.create_1tp_class_model(datadim, len(lab_dict))  # type: ignore
    reg_model = models.create_1tp_reg_model(datadim)  # type: ignore

    run_modes = ["class", "reg"]
    if "class" in run_modes:
        logger.info(f"\nRunning classification model for {ua.data_type}")

        trained_class_model = fit_class_model(
            work_dir,
            class_model,
            ua.data_type,
            ts_train_data,
            train_labs,
            ts_val_data,
            val_labs,
            ua.experiment_name,
        )
        evaluate_class_model(
            trained_class_model,
            ts_test_data,
            test_labs,
            test_reps,
            work_dir,
            yaml_data["scenarios"],
            ua.experiment_name,
            ua.data_type,
            lab_dict,
        )

    if "reg" in run_modes:
        for idx, scenario in enumerate(yaml_data["scenarios"][1:], start=1):
            # print(reg_model.summary())
            logger.info(
                f"{ua.data_type.upper()} Regression {scenario.upper()} TS Data shape (samples, timepoints, alleles/haplotypes): {ts_train_data.shape}"
            )
            train_idxs = np.where(
                (train_s.flatten() > 0.0) & (train_labs[:, idx] == 1)
            )[0]
            val_idxs = np.where((val_s.flatten() > 0.0) & (val_labs[:, idx] == 1))[0]
            test_idxs = np.where((test_s.flatten() > 0.0) & (test_labs[:, idx] == 1))[0]

            mode = "minmax"
            if mode == "log":
                mm_scaler = None
                trvals = -np.log10(train_s[train_idxs])
                vvals = -np.log10(val_s[val_idxs])
                tevals = -np.log10(test_s[test_idxs])
            elif mode == "minmax":
                mm_scaler = scale_sel_coeffs(train_s[train_idxs])
                trvals = mm_scaler.transform(train_s[train_idxs])
                vvals = mm_scaler.transform(val_s[val_idxs])
                tevals = mm_scaler.transform(test_s[test_idxs])
            else:
                mm_scaler = None
                trvals = train_s[train_idxs]
                vvals = val_s[val_idxs]
                tevals = test_s[test_idxs]

            plot = False
            if plot:
                pu.plot_s_vs_freqs(
                    train_s[train_idxs],
                    ts_train_data[train_idxs, -1, int(ts_train_data.shape[-1] / 2)]
                    - ts_train_data[train_idxs, 0, int(ts_train_data.shape[-1] / 2)],
                    scenario,
                    work_dir,
                    ua.experiment_name,
                    mode,
                )

            trained_reg_model = fit_reg_model(
                work_dir,
                reg_model,
                ua.data_type,
                ts_train_data[train_idxs],
                trvals,
                ts_val_data[val_idxs],
                vvals,
                ua.experiment_name + f"_{scenario}_{mode}",
            )
            evaluate_reg_model(
                trained_reg_model,
                ts_test_data[test_idxs],
                test_labs[test_idxs],
                tevals,
                [test_reps[i] for i in list(test_idxs)],
                mm_scaler,
                work_dir,
                yaml_data["scenarios"],
                ua.experiment_name + f"_{scenario}_{mode}",
                ua.data_type,
                lab_dict,
            )



import logging
import math
import multiprocessing as mp
import os
import pickle
import sys
from glob import glob
from itertools import cycle
from random import sample
import allel
import numpy as np
from numpy.random import default_rng
from tqdm import tqdm
from timesweeper.utils import snp_utils as su
from timesweeper.utils.gen_utils import (
    get_rep_id,
    get_scenario_from_filename,
    read_config,
)
from timesweeper.utils.hap_utils import getTSHapFreqs, haps_to_strlist

logging.basicConfig()
logger = logging.getLogger("make_training_feats")
logger.setLevel("INFO")

import warnings

warnings.filterwarnings("error")


def draw_rand_center_offset(max_offset=700):
    rng = default_rng(np.random.seed(int.from_bytes(os.urandom(4), byteorder="little")))
    return int(rng.uniform(-max_offset, max_offset, 1)[0])


def add_missingness(data, m_rate, nan_val=-1):
    rng = default_rng(np.random.seed(int.from_bytes(os.urandom(4), byteorder="little")))
    missing = rng.binomial(1, m_rate, data.shape)
    data[missing == 1] = nan_val

    return data


def check_freq_increase(ts_afs, min_increase=0.25):
    """Quick test to make sure a given example is increasing properly, can be used to filter training."""
    center = int(ts_afs.shape[0] / 2)
    if ts_afs[-1, center] - ts_afs[0, center] >= min_increase:
        return True
    else:
        return False


def subsample_tps(inds_per_tp, og_tps, num_sampled_tps):
    """Subsamples timepoints during VCF by sampling samples list."""
    samples_list = []
    if num_sampled_tps == 1:
        t = og_tps + 1  # VCFs merged samples are 1-indexed, get last
        samples_list.extend([f"{t}:i{i}" for i in range(inds_per_tp)])
    elif num_sampled_tps == 2:
        # This covers the first sampling point
        samples_list.extend([f"i{i}" for i in range(inds_per_tp)])
        t = og_tps + 1  # VCFs merged samples are 1-indexed, get last
        samples_list.extend([f"{t}:i{i}" for i in range(inds_per_tp)])
    else:
        tps = np.linspace(1, og_tps, num_sampled_tps, dtype=int)
        # This covers the first sampling point
        samples_list.extend([f"i{i}" for i in range(inds_per_tp)])
        # This gets the rest
        for t in tps[1:]:
            samples_list.extend([f"{t}:i{i}" for i in range(inds_per_tp)])

    return samples_list


def subsample_inds(inds_per_tp, subsample_size, num_tps):
    """Subsamples individuals from each timepoint to re-use simulations."""
    samples_list = []
    samples_list.extend(
        [f"i{i}" for i in np.random.choice(inds_per_tp, subsample_size, replace=False)]
    )
    for s in range(2, num_tps + 1):
        samples_list.extend(
            [
                f"{s}:i{i}"
                for i in np.random.choice(inds_per_tp, subsample_size, replace=False)
            ]
        )
    return samples_list


def get_window_idxs(center_idx, win_size):
    """
    Gets the win_size number of snps around a central snp.

    Args:
        center_idx (int): Index of the central SNP to use for the window.
        win_size (int): Size of window to use around the SNP, optimally odd number.

    Returns:
        list: Indices of all SNPs to grab for the feature matrix.
    """
    half_window = math.floor(win_size / 2)
    return list(range(center_idx - half_window, center_idx + half_window + 1))


def prep_ts_aft(genos, samp_sizes):
    """
    Iterates through timepoints and creates MAF feature matrices.

    Args:
        genos (allel.GenotypeArray): Genotype array containing all timepoints.
        samp_sizes (list[int]): Number of chromosomes sampled at each timepoint.

    Returns:
        np.arr: MAF array to use for predictions. Shape is (timepoints, MAF).
    """
    # Prep genos into time-series format and calculate Maft
    ts_genos = su.split_arr(genos, samp_sizes)

    min_alleles, first_genos, last_genos = su.get_vel_minor_alleles(
        ts_genos, np.max(genos)
    )

    ts_maft = []
    for timepoint in ts_genos:
        _genos = []
        _genotypes = allel.GenotypeArray(timepoint).count_alleles(
            max_allele=min_alleles.max()
        )

        for snp, min_allele_idx in zip(_genotypes, min_alleles):
            maf = su.calc_maft(snp, min_allele_idx)
            _genos.append(maf)

        ts_maft.append(_genos)

    return np.stack(ts_maft)


def get_aft_central_window(
    snps, genos, samp_sizes, win_size, missingness, mut_types, offset
):
    """
    Check for whether a non-control mutation type is present. If not, return the central-most mutation.

    Args:
        snps (list[tup(chrom, pos,  mut)]): Tuples of information for each SNP.
        genos (allel.GenotypeArray): Genotypes of all samples.
        samp_sizes (list[int]): Number of chromosomes sampled at each timepoint.
        win_size (int): Number of SNPs to use for each prediction. Needs to match how NN was trained.
        scenario (str): Entry from the scenarios config option.
        missingness (float): Parameter of binomial distribution to pull missingness from.
        mut_types (list[int]): List of mutation types that are not considered the "control" case.
    Returns:
        np.arr: The central-most window, either based on mutation type or closest to half size of chrom.
        float: Selection coefficient.
    """
    ts_aft = prep_ts_aft(genos, samp_sizes)

    buffer = int(win_size / 2)
    centers = range(buffer, len(snps) - buffer)

    center_idx = int(len(centers) / 2)

    try:
        for center in centers:
            if snps[center][2] in mut_types:
                center_idx = center
                break
            else:
                pass
    except:
        pass

    sel_coeff = snps[center_idx][3]

    if offset:
        rng = default_rng(
            np.random.seed(int.from_bytes(os.urandom(4), byteorder="little"))
        )
        if rng.uniform(0, 3, 1)[0] > 2:
            rand_offset = draw_rand_center_offset()
            center_idx += rand_offset
        else:
            rand_offset = 0
    else:
        rand_offset = 0

    win_idxs = get_window_idxs(center_idx, win_size)

    window = ts_aft[:, win_idxs]
    center_aft = window

    missing_center_aft = add_missingness(
        center_aft, m_rate=missingness
    )  # If no missingness, will just return

    return missing_center_aft, sel_coeff, rand_offset


def get_hft_central_window(snps, haps, samp_sizes, win_size, mut_types, offset):
    """
    Iterates through windows of MAF time-series matrix and gets the central window.
    Does not have as many utility functions as AFT such as missingness and variable sorting methods.

    Args:
        snps (list[tup(chrom, pos,  mut)]): Tuples of information for each SNP.
        haps (np.arr): Haplotypes of all samples.
        samp_sizes (list[int]): Number of chromosomes sampled at each timepoint.
        win_size (int): Number of SNPs to use for each prediction. Needs to match how NN was trained.
    Returns:
        np.arr: The central-most window, either based on mutation type or closest to half size of chrom.
    """
    buffer = int(win_size / 2)
    centers = range(buffer, len(snps) - buffer)

    center_idx = int(len(centers) / 2)

    try:
        for center in centers:
            if snps[center][2] in mut_types:
                center_idx = center
                break
            else:
                pass
    except:
        pass

    sel_coeff = snps[center_idx][3]

    if offset:
        rng = default_rng(
            np.random.seed(int.from_bytes(os.urandom(4), byteorder="little"))
        )
        if rng.uniform(1, 3, 1)[0] > 2:
            rand_offset = draw_rand_center_offset()
            center_idx += rand_offset

        else:
            rand_offset = 0
    else:
        rand_offset = 0

    win_idxs = get_window_idxs(center_idx, win_size)
    window = np.swapaxes(haps[win_idxs, :], 0, 1)
    str_window = haps_to_strlist(window)
    central_hfs = getTSHapFreqs(str_window, samp_sizes)

    return central_hfs, sel_coeff, rand_offset


def aft_worker(
    in_vcf,
    mut_types,
    scenarios,
    samp_sizes,
    samps_list,
    win_size,
    offset,
    missingness,
    verbose=False,
):
    benchmark = True  # Want to get all the info we can from sims in training
    try:
        id = get_rep_id(in_vcf)
        scenario = get_scenario_from_filename(in_vcf, scenarios)

        vcf = su.read_vcf(in_vcf, samps_list, benchmark)
        genos, snps = su.vcf_to_genos(vcf, benchmark)

        central_aft, sel_coeff, rand_offset = get_aft_central_window(
            snps, genos, samp_sizes, win_size, missingness, mut_types, offset
        )

        if "neut" not in scenario.lower() and sel_coeff == 0.0:
            raise Exception

        return id, scenario, central_aft, sel_coeff, rand_offset

    except UserWarning as Ue:
        print(Ue)
        return None

    except Exception as e:
        if verbose:
            logger.warning(f"Could not process {in_vcf}")
            logger.warning(f"Exception: {e}")
            sys.stdout.flush()
            sys.stderr.flush()
        return None


def hft_worker(
    in_vcf,
    mut_types,
    scenarios,
    samp_sizes,
    samps_list,
    win_size,
    offset,
    ploidy=2,
    verbose=False,
):
    benchmark = True
    try:
        id = get_rep_id(in_vcf)
        scenario = get_scenario_from_filename(in_vcf, scenarios)

        vcf = su.read_vcf(in_vcf, samps_list, benchmark)
        haps, snps = su.vcf_to_haps(vcf, benchmark)

        central_hft, sel_coeff, rand_offset = get_hft_central_window(
            snps, haps, [ploidy * i for i in samp_sizes], win_size, mut_types, offset
        )

        return id, scenario, central_hft, sel_coeff, rand_offset

    except UserWarning as Ue:
        # print(Ue)
        return None

    except Exception as e:
        if verbose:
            logger.warning(f"Could not process {in_vcf}")
            logger.warning(f"Exception: {e}")
            sys.stdout.flush()
            sys.stderr.flush()
        return None


def main(ua):
    yaml_data = read_config(ua.yaml_file)
    scenarios, mut_types, work_dir, samp_sizes, ploidy, win_size, threads = (
        yaml_data["scenarios"],
        yaml_data["mut types"],
        yaml_data["work dir"],
        yaml_data["sample sizes"],
        yaml_data["ploidy"],
        yaml_data["win_size"],
        ua.threads,
    )
    if ua.subsample_inds:
        if ua.subsample_tps:
            logger.error("Can't subsample both timepoints and individuals.")

        samps_list = subsample_inds(samp_sizes[0], ua.subsample_inds, len(samp_sizes))
    else:
        samps_list = None

    if ua.subsample_tps:
        if not ua.og_tps:
            logger.error(
                "Must provide original simulation number of tps in order to properly subset."
            )

        samps_list = subsample_tps(samp_sizes[0], ua.og_tps, ua.subsample_tps)

    if ua.allow_shoulders:
        offset = int(ua.allow_shoulders)
    else:
        offset = 0

    filelist = glob(f"{work_dir}/vcfs/*/*/merged.vcf", recursive=True)

    aft_work_args = zip(
        filelist,
        cycle([mut_types]),
        cycle([scenarios]),
        cycle([samp_sizes]),
        cycle([samps_list]),
        cycle([win_size]),
        cycle([offset]),
        cycle([ua.missingness]),
        cycle([ua.verbose]),
    )
    hft_work_args = zip(
        filelist,
        cycle([mut_types]),
        cycle([scenarios]),
        cycle([samp_sizes]),
        cycle([samps_list]),
        cycle([win_size]),
        cycle([offset]),
        cycle([int(ploidy)]),
        cycle([ua.verbose]),
    )

    debug = False
    if debug:
        aft_work_res = [aft_worker(*i) for i in aft_work_args]
        hft_work_res = [hft_worker(*i) for i in hft_work_args]
    else:
        pool = mp.Pool(threads)
        if ua.no_progress:
            aft_work_res = pool.starmap(aft_worker, aft_work_args, chunksize=4,)

            if ua.hft:
                hft_work_res = pool.starmap(hft_worker, hft_work_args, chunksize=4,)

            pool.close()
        else:
            aft_work_res = pool.starmap(
                aft_worker,
                tqdm(
                    aft_work_args,
                    desc="Formatting AFT training data",
                    total=len(filelist),
                ),
                chunksize=4,
            )
            if ua.hft:
                hft_work_res = pool.starmap(
                    hft_worker,
                    tqdm(
                        hft_work_args,
                        desc="Formatting HFT training data",
                        total=len(filelist),
                    ),
                    chunksize=4,
                )
            pool.close()

    pickle_dict = {}
    for s in scenarios:
        pickle_dict[s] = {}

    for res in aft_work_res:
        if res:
            rep, scenario, aft, s, off = res

            pickle_dict[scenario][rep] = {}
            pickle_dict[scenario][rep]["aft"] = aft
            pickle_dict[scenario][rep]["sel_coeff"] = s
            pickle_dict[scenario][rep]["center_offset"] = off

    if ua.hft:
        for res in hft_work_res:
            try:
                if res:
                    rep, scenario, hft, s, off = res
                    if rep in pickle_dict[scenario].keys():
                        pickle_dict[scenario][rep]["hft"] = hft
                        pickle_dict[scenario][rep]["sel_coeff"] = s
                        pickle_dict[scenario][rep]["center_offset"] = off

            except KeyError as e:
                print(e)
                print(res)
                pass

    with open(ua.outfile, "wb") as outfile:
        pickle.dump(pickle_dict, outfile)



from tensorflow.keras import layers
from tensorflow.keras.models import Model

# fmt: off
def create_TS_class_model(datadim, n_class):
    """
    Returns:
        Model: Keras compiled model.
    """
    model_in = layers.Input(datadim)
    h = layers.Conv1D(64, 3, activation="relu", padding="same")(model_in)
    h = layers.Conv1D(64, 3, activation="relu", padding="same")(h)
    h = layers.MaxPooling1D(pool_size=3, padding="same")(h)
    h = layers.Dropout(0.15)(h)
    h = layers.Flatten()(h)

    h = layers.Dense(512, activation="relu")(h)
    h = layers.Dropout(0.2)(h)        
    h = layers.Dense(264, activation="relu")(h)
    h = layers.Dropout(0.2)(h)
    h = layers.Dense(128, activation="relu")(h)
    h = layers.Dropout(0.1)(h)
    class_output = layers.Dense(n_class, activation="softmax", name="class_output")(h)

    model = Model(inputs=[model_in], outputs=[class_output], name="Timesweeper_Class")
    model.compile(
        loss={"class_output":"categorical_crossentropy"},
        optimizer="adam",
        metrics={"class_output": "accuracy"},
    )

    return model

def create_TS_reg_model(datadim):
    """
    Returns:
        Model: Keras compiled model.
    """
    model_in = layers.Input(datadim)
    h = layers.Conv1D(64, 3, activation="relu", padding="same")(model_in)
    h = layers.Conv1D(64, 3, activation="relu", padding="same")(h)
    h = layers.MaxPooling1D(pool_size=3, padding="same")(h)
    h = layers.Dropout(0.15)(h)
    h = layers.Flatten()(h)

    h = layers.Dense(264, activation="relu")(h)
    h = layers.Dropout(0.2)(h)        
    h = layers.Dense(264, activation="relu")(h)
    h = layers.Dropout(0.2)(h)
    h = layers.Dense(128, activation="relu")(h)
    h = layers.Dropout(0.1)(h)
    reg_output = layers.Dense(1, activation="linear", name="reg_output")(h)

    model = Model(inputs=[model_in], outputs=[reg_output], name="Timesweeper_Reg")
    model.compile(
        loss={"reg_output":"mse"},
        optimizer="adam",
        metrics={"reg_output": "mse"},
    )

    return model


# fmt: off
def create_2D_TS_class_model(datadim, n_class):
    """
    Returns:
        Model: Keras compiled model.
    """
    model_in = layers.Input(datadim)
    h = layers.Conv2D(64, 3, activation="relu", padding="same")(model_in)
    h = layers.Conv2D(64, 3, activation="relu", padding="same")(h)
    h = layers.MaxPooling2D(pool_size=3, padding="same")(h)
    h = layers.Dropout(0.15)(h)
    h = layers.Flatten()(h)

    h = layers.Dense(512, activation="relu")(h)
    h = layers.Dropout(0.2)(h)        
    h = layers.Dense(264, activation="relu")(h)
    h = layers.Dropout(0.2)(h)
    h = layers.Dense(128, activation="relu")(h)
    h = layers.Dropout(0.1)(h)
    class_output = layers.Dense(n_class, activation="softmax", name="class_output")(h)

    model = Model(inputs=[model_in], outputs=[class_output], name="2DTimesweeper_Class")
    model.compile(
        loss={"class_output":"categorical_crossentropy"},
        optimizer="adam",
        metrics={"class_output": "accuracy"},
    )

    return model

def create_2D_TS_reg_model(datadim):
    """
    Returns:
        Model: Keras compiled model.
    """
    model_in = layers.Input(datadim)
    
    h = layers.Conv2D(64, 3, activation="relu", padding="same")(model_in)
    h = layers.Conv2D(64, 3, activation="relu", padding="same")(h)
    h = layers.MaxPooling2D(pool_size=3, padding="same")(h)
    h = layers.Dropout(0.15)(h)
    h = layers.Flatten()(h)

    h = layers.Dense(264, activation="relu")(h)
    h = layers.Dropout(0.2)(h)        
    h = layers.Dense(264, activation="relu")(h)
    h = layers.Dropout(0.2)(h)
    h = layers.Dense(128, activation="relu")(h)
    h = layers.Dropout(0.1)(h)
    reg_output = layers.Dense(1, activation="relu", name="reg_output")(h)

    model = Model(inputs=[model_in], outputs=[reg_output], name="2DTimesweeper_Reg")
    model.compile(
        loss={"reg_output":"mse"},
        optimizer="adam",
        metrics={"reg_output": "mse"},
    )

    return model


# fmt: off
def create_big_TS_class_model(datadim, n_class):
    """
    Returns:
        Model: Keras compiled model.
    """
    model_in = layers.Input(datadim)
    h = layers.Conv1D(1028, 3, activation="relu", padding="same")(model_in)
    h = layers.Conv1D(1028, 3, activation="relu", padding="same")(h)
    h = layers.MaxPooling1D(pool_size=3, padding="same")(h)
    h = layers.Dropout(0.15)(h)
    h = layers.Flatten()(h)

    h = layers.Dense(1028, activation="relu")(h)
    h = layers.Dropout(0.2)(h)        
    h = layers.Dense(512, activation="relu")(h)
    h = layers.Dropout(0.2)(h)
    h = layers.Dense(128, activation="relu")(h)
    h = layers.Dropout(0.1)(h)
    class_output = layers.Dense(n_class, activation="softmax", name="class_output")(h)

    model = Model(inputs=[model_in], outputs=[class_output], name="Big_Timesweeper_Class")
    model.compile(
        loss={"class_output":"categorical_crossentropy"},
        optimizer="adam",
        metrics={"class_output": "accuracy"},
    )

    return model

def create_big_TS_reg_model(datadim):
    """
    Returns:
        Model: Keras compiled model.
    """
    model_in = layers.Input(datadim)
    h = layers.Conv1D(1028, 3, activation="relu", padding="same")(model_in)
    h = layers.Conv1D(1028, 3, activation="relu", padding="same")(h)
    h = layers.MaxPooling1D(pool_size=3, padding="same")(h)
    h = layers.Dropout(0.15)(h)
    h = layers.Flatten()(h)

    h = layers.Dense(1028, activation="relu")(h)
    h = layers.Dropout(0.2)(h)        
    h = layers.Dense(512, activation="relu")(h)
    h = layers.Dropout(0.2)(h)
    h = layers.Dense(128, activation="relu")(h)
    h = layers.Dropout(0.1)(h)
    reg_output = layers.Dense(1, activation="relu", name="reg_output")(h)

    model = Model(inputs=[model_in], outputs=[reg_output], name="Big_Timesweeper_Reg")
    model.compile(
        loss={"reg_output":"mse"},
        optimizer="adam",
        metrics={"reg_output": "mse"},
    )

    return model


# fmt: off
def create_1tp_class_model(datadim, n_class):
    """
    Returns:
        Model: Keras compiled model.
    """
    model_in = layers.Input(datadim)

    h = layers.Dense(264, activation="relu")(model_in)
    h = layers.Dropout(0.2)(h)        
    h = layers.Dense(128, activation="relu")(h)
    h = layers.Dropout(0.2)(h)
    h = layers.Dense(56, activation="relu")(h)
    h = layers.Dropout(0.1)(h)
    class_output = layers.Dense(n_class, activation="softmax", name="class_output")(h)

    model = Model(inputs=[model_in], outputs=[class_output], name="Timesweeper_Class")
    model.compile(
        loss={"class_output":"categorical_crossentropy"},
        optimizer="adam",
        metrics={"class_output": "accuracy"},
    )

    return model


def create_1tp_reg_model(datadim):
    """
    Returns:
        Model: Keras compiled model.
    """
    model_in = layers.Input(datadim)
    h = layers.Dense(264, activation="relu")(model_in)
    h = layers.Dropout(0.2)(h)        
    h = layers.Dense(128, activation="relu")(h)
    h = layers.Dropout(0.2)(h)
    h = layers.Dense(56, activation="relu")(h)
    h = layers.Dropout(0.1)(h)

    reg_output = layers.Dense(1, activation="relu", name="reg_output")(h)

    model = Model(inputs=[model_in], outputs=[reg_output], name="Timesweeper_Reg")
    model.compile(
        loss={"reg_output":"mse"},
        optimizer="adam",
        metrics={"reg_output": "mse"},
    )

    return model
# fmt: on


def create_rnn_class_model(datadim, n_class):
    model_in = layers.Input(datadim)
    h = layers.LSTM(100, input_shape=datadim, activation="tanh")(model_in)
    h = layers.Dropout(0.1)(h)

    class_output = layers.Dense(n_class, activation="softmax", name="class_output")(h)

    model = Model(inputs=[model_in], outputs=[class_output], name="Timesweeper_Class")
    model.compile(
        loss={"class_output": "categorical_crossentropy"},
        optimizer="adam",
        metrics={"class_output": "accuracy"},
    )

    return model


def create_rnn_reg_model(datadim):
    model_in = layers.Input(datadim)
    h = layers.LSTM(100, input_shape=datadim, activation="tanh")(model_in)
    h = layers.Dropout(0.1)(h)

    reg_output = layers.Dense(1, activation="relu", name="reg_output")(h)

    model = Model(inputs=[model_in], outputs=[reg_output], name="RNN_Reg")
    model.compile(
        loss={"reg_output": "mse"}, optimizer="adam", metrics={"reg_output": "mse"},
    )

    return model


def transformer_encoder(inputs, head_size, num_heads, ff_dim, dropout=0):
    # Attention and Normalization
    x = layers.MultiHeadAttention(
        key_dim=head_size, num_heads=num_heads, dropout=dropout
    )(inputs, inputs)
    x = layers.Dropout(dropout)(x)
    x = layers.LayerNormalization(epsilon=1e-6)(x)
    res = x + inputs

    # Feed Forward Part
    x = layers.Conv1D(filters=ff_dim, kernel_size=1, activation="relu")(res)
    x = layers.Dropout(dropout)(x)
    x = layers.Conv1D(filters=inputs.shape[-1], kernel_size=1)(x)
    x = layers.LayerNormalization(epsilon=1e-6)(x)

    return x + res


def create_transformer_reg_model(
    input_shape,
    head_size,
    num_heads,
    ff_dim,
    num_transformer_blocks,
    mlp_units,
    dropout=0,
    mlp_dropout=0,
):
    inputs = layers.Input(shape=input_shape)
    x = inputs
    for _ in range(num_transformer_blocks):
        x = transformer_encoder(x, head_size, num_heads, ff_dim, dropout)

    x = layers.GlobalAveragePooling1D(data_format="channels_last")(x)
    for dim in mlp_units:
        x = layers.Dense(dim, activation="relu")(x)
        x = layers.Dropout(mlp_dropout)(x)
    outputs = layers.Dense(1, activation="relu")(x)

    model = Model(inputs, outputs, name="Timesweeper_Transformer_Reg")

    model.compile(
        loss="mse", optimizer="adam", metrics="mse",
    )

    return model


def create_transformer_class_model(
    input_shape,
    head_size,
    num_heads,
    ff_dim,
    num_transformer_blocks,
    mlp_units,
    dropout=0,
    mlp_dropout=0,
    n_class=3,
):
    inputs = layers.Input(shape=input_shape)
    x = inputs
    for _ in range(num_transformer_blocks):
        x = transformer_encoder(x, head_size, num_heads, ff_dim, dropout)

    x = layers.GlobalAveragePooling1D(data_format="channels_last")(x)
    for dim in mlp_units:
        x = layers.Dense(dim, activation="relu")(x)
        x = layers.Dropout(mlp_dropout)(x)
    outputs = layers.Dense(n_class, activation="softmax")(x)

    model = Model(inputs, outputs, name="Timesweeper_Transformer_Class")

    model.compile(
        loss="categorical_crossentropy", optimizer="adam", metrics="accuracy",
    )

    return model



import math
import os

import numpy as np
import pickle as pkl
import pandas as pd
from tensorflow.keras.models import load_model
from tqdm import tqdm

from timesweeper.make_training_features import prep_ts_aft, get_window_idxs
from timesweeper.utils import snp_utils as su
from timesweeper.utils.gen_utils import read_config, get_logger
from timesweeper.utils import hap_utils as hu

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

logger = get_logger("find_sweeps")


def write_preds(results_dict, outfile, scaler, benchmark):
    """
    Writes NN predictions to file.

    Args:
        results_dict (dict): SNP NN prediction scores and window edges.
        outfile (str): File to write results to.
    """
    with open(scaler, "rb") as ifile:
        scaler = pkl.load(ifile)

    lab_dict = {0: "Neut", 1: "SSV", 2: "SDN"}
    if benchmark:
        chroms, bps, mut_type, true_sel_coeff = zip(*results_dict.keys())
    else:
        chroms, bps = zip(*results_dict.keys())

    neut_scores = [i[0][0] for i in results_dict.values()]
    sdn_scores = [i[0][1] for i in results_dict.values()]
    ssv_scores = [i[0][2] for i in results_dict.values()]
    sdn_preds = [i[1] for i in results_dict.values()]
    ssv_preds = [i[2] for i in results_dict.values()]
    left_edges = [i[3] for i in results_dict.values()]
    right_edges = [i[4] for i in results_dict.values()]
    classes = [lab_dict[np.argmax(i[0])] for i in results_dict.values()]

    sdn_s = scaler.inverse_transform(sdn_preds).squeeze().flatten()
    ssv_s = scaler.inverse_transform(ssv_preds).squeeze().flatten()

    if benchmark:
        predictions = pd.DataFrame(
            {
                "Chrom": chroms,
                "BP": bps,
                "Mut_Type": mut_type,
                "Class": classes,
                "Sweep_Prob": [i + j for i, j in zip(sdn_scores, ssv_scores)],
                "True_Sel_Coeff": true_sel_coeff,
                "SDN_s_pred": sdn_s,
                "SSV_s_pred": ssv_s,
                "Neut_Prob": neut_scores,
                "SDN_Prob": sdn_scores,
                "SSV_Prob": ssv_scores,
                "Win_Start": left_edges,
                "Win_End": right_edges,
            }
        )
    else:
        predictions = pd.DataFrame(
            {
                "Chrom": chroms,
                "BP": bps,
                "Class": classes,
                "Sweep_Prob": [i + j for i, j in zip(sdn_scores, ssv_scores)],
                "SDN_s_pred": sdn_s,
                "SSV_s_pred": ssv_s,
                "Neut_Prob": neut_scores,
                "SDN_Prob": sdn_scores,
                "SSV_Prob": ssv_scores,
                "Win_Start": left_edges,
                "Win_End": right_edges,
            }
        )
        predictions = predictions[predictions["Neut_Prob"] < 0.5]

    predictions.sort_values(["Chrom", "BP"], inplace=True)

    predictions.to_csv(outfile, header=True, index=False, sep="\t")

    bed_df = predictions[["Chrom", "Win_Start", "Win_End", "BP"]]
    bed_df.to_csv(
        outfile.replace(".csv", ".bed"), mode="a", header=False, index=False, sep="\t"
    )


def run_aft_windows(
    snps, genos, samp_sizes, win_size, class_model, sdn_model, ssv_model
):
    """
    Iterates through windows of MAF time-series matrix and predicts using NN.

    Args:
        snps (list[tup(chrom, pos,  mut)]): Tuples of information for each SNP. Contains mut only if benchmarking == True.
        genos (allel.GenotypeArray): Genotypes of all samples.
        samp_sizes (list[int]): Number of chromosomes sampled at each timepoint.
        win_size (int): Number of SNPs to use for each prediction. Needs to match how NN was trained.
        model (Keras.model): Keras model to use for prediction.

    Returns:
        dict: Prediction values in the form of dict[snps[center]]
        np.arr: the central-most window, either based on mutation type or closest to half size of chrom.
    """
    ts_aft = prep_ts_aft(genos, samp_sizes)

    # Iterate over SNP windows and predict
    buffer = math.floor(win_size / 2)

    centers = range(buffer, len(snps) - buffer)
    left_edges = []
    right_edges = []
    data = []
    for center in tqdm(centers, desc="Predicting on AFT windows"):
        try:
            win_idxs = get_window_idxs(center, win_size)
            window = ts_aft[:, win_idxs]
            data.append(window)
            left_edges.append(snps[win_idxs[0]][1])
            right_edges.append(snps[win_idxs[-1]][1])

        except Exception as e:
            logger.warning(f"Center {snps[center]} raised error {e}")

    class_probs = class_model.predict(np.stack(data))
    sdn_sel_preds = sdn_model.predict(np.stack(data))
    ssv_sel_preds = ssv_model.predict(np.stack(data))

    results_dict = {}
    for center, class_probs, sd, sv, l_e, r_e in zip(
        centers, class_probs, sdn_sel_preds, ssv_sel_preds, left_edges, right_edges
    ):
        results_dict[snps[center]] = (
            class_probs,
            sd,
            sv,
            l_e,
            r_e,
        )

    return results_dict


def run_hft_windows(
    snps, haps, ploidy, samp_sizes, win_size, class_model, sdn_model, ssv_model
):
    """
    Iterates through windows of MAF time-series matrix and predicts using NN.
    Args:
        snps (list[tup(chrom, pos,  mut)]): Tuples of information for each SNP. Contains mut only if benchmarking == True.
        haps (np.arr): Haplotypes of all samples.
        samp_sizes (list[int]): Number of chromosomes sampled at each timepoint.
        win_size (int): Number of SNPs to use for each prediction. Needs to match how NN was trained.
        model (Keras.model): Keras model to use for prediction.
    Returns:
        dict: Prediction values in the form of dict[snps[center]]
        np.arr: the central-most window, either based on mutation type or closest to half size of chrom.
    """
    results_dict = {}
    buffer = math.floor(win_size / 2)
    centers = range(buffer, len(snps) - buffer)
    left_edges = []
    right_edges = []
    data = []
    for center in tqdm(centers, desc="Predicting on HFT windows"):
        # try:
        win_idxs = get_window_idxs(center, win_size)
        window = np.swapaxes(haps[win_idxs, :], 0, 1)
        str_window = hu.haps_to_strlist(window)
        hft = hu.getTSHapFreqs(str_window, [i * ploidy for i in samp_sizes])
        data.append(hft)
        left_edges.append(snps[win_idxs[0]][1])
        right_edges.append(snps[win_idxs[-1]][1])

        # except Exception as e:
        #    logger.warning(f"Center {snps[center]} raised error {e}")
        #    continue

    class_probs = class_model.predict(np.stack(data))
    sdn_sel_preds = sdn_model.predict(np.stack(data))
    ssv_sel_preds = ssv_model.predict(np.stack(data))

    results_dict = {}
    for center, class_probs, sd, sv, l_e, r_e in zip(
        centers, class_probs, sdn_sel_preds, ssv_sel_preds, left_edges, right_edges
    ):
        results_dict[snps[center]] = (
            class_probs,
            sd,
            sv,
            l_e,
            r_e,
        )

    return results_dict


def get_window_idxs(center_idx, win_size):
    """
    Gets the win_size number of snps around a central snp.

    Args:
        center_idx (int): Index of the central SNP to use for the window.
        win_size (int): Size of window to use around the SNP, optimally odd number.

    Returns:
        list: Indices of all SNPs to grab for the feature matrix.
    """
    half_window = math.floor(win_size / 2)
    return list(range(center_idx - half_window, center_idx + half_window + 1))


def load_nn(model_path, summary=False):
    """
    Loads the trained Keras network.

    Args:
        model_path (str): Path to Keras model.
        summary (bool, optional): Whether to print out model summary or not. Defaults to False.

    Returns:
        Keras.model: Trained Keras model to use for prediction.
    """
    model = load_model(model_path)
    if summary:
        print(model.summary())

    return model


def main(ua):
    yaml_data = read_config(ua.yaml_file)
    samp_sizes = yaml_data["sample sizes"]
    ploidy = yaml_data["ploidy"]
    win_size = yaml_data["win_size"]

    class_aft_model = load_nn(ua.aft_class_model)
    if "ssv" in ua.aft_reg_model:
        aft_ssv_model = load_model(ua.aft_reg_model)
        aft_sdn_model = load_model(str(ua.aft_reg_model).replace("ssv", "sdn"))
    elif "sdn" in ua.aft_reg_model:
        aft_ssv_model = load_model(str(ua.aft_reg_model).replace("ssv", "sdn"))
        aft_sdn_model = load_model(ua.aft_reg_model)

    if ua.hft_class_model:
        class_hft_model = load_nn(ua.hft_class_model)
        if "ssv" in ua.hft_reg_model:
            hft_ssv_model = load_model(ua.hft_reg_model)
            hft_sdn_model = load_model(str(ua.hft_reg_model).replace("ssv", "sdn"))
        elif "sdn" in ua.hft_reg_model:
            hft_ssv_model = load_model(str(ua.hft_reg_model).replace("ssv", "sdn"))
            hft_sdn_model = load_model(ua.hft_reg_model)

    outfile = ua.outfile

    # Chunk and iterate for NN predictions to not take up too much space
    vcf_iter = su.get_vcf_iter(ua.input_vcf, ua.benchmark)
    for chunk_idx, chunk in enumerate(vcf_iter):
        chunk = chunk[0]  # Why you gotta do me like that, skallel?
        logger.info(f"Processing VCF chunk {chunk_idx}")

        # aft
        # try:
        genos, snps = su.vcf_to_genos(chunk, ua.benchmark)
        aft_predictions = run_aft_windows(
            snps,
            genos,
            samp_sizes,
            win_size,
            class_aft_model,
            aft_sdn_model,
            aft_ssv_model,
        )
        write_preds(aft_predictions, f"{outfile}_aft.csv", ua.scalar, ua.benchmark)

        # except Exception as e:
        #    logger.error(f"Cannot process chunk {chunk_idx} using AFT due to {e}")

        # hft
        if ua.hft_class_model:
            # try:
            haps, snps = su.vcf_to_haps(chunk, ua.benchmark)
            hft_predictions = run_hft_windows(
                snps,
                haps,
                ploidy,
                samp_sizes,
                win_size,
                class_hft_model,
                hft_sdn_model,
                hft_ssv_model,
            )
            write_preds(hft_predictions, f"{outfile}_hft.csv", ua.scalar, ua.benchmark)

            # except Exception as e:
            #    logger.error(f"Cannot process chunk {chunk_idx} using HFT due to {e}")



import logging
import os
import pickle
import random

import numpy as np
import pandas as pd
import tensorflow as tf
from sklearn.metrics import confusion_matrix, mean_absolute_error
from sklearn.model_selection import train_test_split
from sklearn.utils import compute_class_weight
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint
from tensorflow.keras.layers import Conv1D, Dense, Dropout, Flatten, Input, MaxPooling1D
from tensorflow.keras.models import Model, save_model
from tensorflow.keras.utils import to_categorical
from tensorflow.keras.utils import plot_model

from timesweeper.plotting import plotting_utils as pu
from timesweeper.utils.gen_utils import read_config

from sklearn.model_selection import GridSearchCV
from tensorflow.keras.wrappers.scikit_learn import KerasClassifier


os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

logging.basicConfig()
logger = logging.getLogger("nets")
logger.setLevel("INFO")

seed = 42
random.seed(seed)
np.random.seed(seed)
tf.random.set_seed(seed)



grid = ParameterGrid([{'kernel_size': [16, 32, 64, 128, 256]},
                      {'kernel': ['rbf'], 'gamma': [1, 10]}]

list(ParameterGrid(grid)) == [{'kernel': 'linear'},
                              {'kernel': 'rbf', 'gamma': 1},
                              {'kernel': 'rbf', 'gamma': 10}]



def get_mlp_model(datadim, n_class,
                  kernel_sizeLayerer_a=64,
                  kernel_sizeLayerer_b=64,
	              dropoutCovLayer_a=0.15,
                  learnRate=0.01,
                  add_one_layer=False,
                  add_two_layer=False,
                  dropoutFullLayer_a=0.2,
                  dropoutFullLayer_b=0.2,
                  dropoutFullLayer_c=0.1):
    """
        Create a CNN model based on the hypter parameter input for the whole sample

    Args:
          kernel_sizeLayerer_a (int): the kernel size of the first convolutiona layer in the first convlutional block.
          kernel_sizeLayerer_b (int): the kernel size of the second convolutiona layer in the first convlutional block.
          dropoutCovLayer_a (float): the dropout rate in the first convlutional block.
          learnRate (float): the learning rate of the CNN.
          add_one_layer (boolean): if true, add one additional convlutional block.
          add_two_layer (boolean): if true, add a second additional convlutional block.
          dropoutFullLayer_a (float): the dropout rate of the first layer in the fully connnected block.
          dropoutFullLayer_b (float): the dropout rate of the second layer in the fully connnected block.
          dropoutFullLayer_c (float): the dropout rate of the third layer in the fully connnected block.

    Returns:
        kera.classifier: a cnn fitted to the input hyperparameter.
    """
    ## Convolution Layer One
    model_in = Input(datadim)
    h = Conv1D(kernel_sizeLayerer_a, 3, activation="relu", padding="same")(model_in)
    h = Conv1D(kernel_sizeLayerer_b, 3, activation="relu", padding="same")(h)
    h = MaxPooling1D(pool_size=3, padding="same")(h)
    h = Dropout(dropoutCovLayer_a)(h)
    h = Flatten()(h)

    ## Convolution Layer Two if add_one_layer
    if add_one_layer:
        h = Conv1D(kernel_sizeLayerer_a, 3, activation="relu", padding="same")(h)
        h = Conv1D(kernel_sizeLayerer_b, 3, activation="relu", padding="same")(h)
        h = MaxPooling1D(pool_size=3, padding="same")(h)
        h = Dropout(dropoutCovLayer_a)(h)
        h = Flatten()(h)

    ## Convolution Layer Three if add_two_layer
    if add_two_layer:
        h = Conv1D(kernel_sizeLayerer_a, 3, activation="relu", padding="same")(h)
        h = Conv1D(kernel_sizeLayerer_b, 3, activation="relu", padding="same")(h)
        h = MaxPooling1D(pool_size=3, padding="same")(h)
        h = Dropout(dropoutCovLayer_a)(h)
        h = Flatten()(h)

    ## Fully Connected Layer
    h = Dense(264, activation="relu")(h)
    h = Dropout(dropoutFullLayer_a)(h)
    h = Dense(264, activation="relu")(h)
    h = Dropout(dropoutFullLayer_b)(h)
    h = Dense(128, activation="relu")(h)
    h = Dropout(dropoutFullLayer_c)(h)

    reg_output = Dense(1, activation="relu", name="reg_output")(h)
    class_output = Dense(n_class, activation="softmax", name="class_output")(h)

    model = Model(inputs=[model_in], outputs=[class_output, reg_output], name="Timesweeper")
    model.compile(
        loss={"class_output":"categorical_crossentropy", "reg_output":"mse"},
        optimizer="adam",
        metrics={"class_output": "accuracy", "reg_output": "mse"},
    )

    return model



def main(ua):
    if ua.config_format == "yaml":
        yaml_data = read_config(ua.yaml_file)
        work_dir = yaml_data["work dir"]

    elif ua.config_format == "cli":
        work_dir = ua.work_dir

    # Collect all the data
    logger.info("Starting training process.")
    if ua.hft:
        type_list = ["aft", "hft"]
    else:
        type_list = ["aft"]
    for data_type in type_list:
        ids, ts_data, sweep_types, sel_coeffs = get_data(ua.training_data, data_type)
        lab_dict = {str_id: int_id for int_id, str_id in enumerate(sweep_types)}

        # Convert to numerical ohe IDs
        num_ids = np.array([lab_dict[lab] for lab in ids])
        ohe_ids = to_categorical(
            np.array([lab_dict[lab] for lab in ids]), len(set(ids))
        )

        class_weights = dict(
            enumerate(
                compute_class_weight(
                    class_weight="balanced", classes=np.unique(num_ids), y=num_ids
                )
            )
        )
        print(f"Class weights: {class_weights}")

        if data_type == "aft":
            # Needs to be in correct dims order for Conv1D layer
            datadim = ts_data.shape[1:]
            logger.info(
                f"TS Data shape (samples, timepoints, alleles): {ts_data.shape}"
            )
        else:
            logger.info(f"TS Data shape (samples, timepoints, haps): {ts_data.shape}")
            datadim = ts_data.shape[1:]
            logger.info(f"{len(ts_data)} samples in dataset.")

        logger.info("Splitting Partition")
        (
            ts_train_data,
            ts_val_data,
            ts_test_data,
            train_labs,
            val_labs,
            test_labs,
            train_s,
            val_s,
            test_s,
        ) = split_partitions(ts_data, ohe_ids, sel_coeffs)

        # Time-series model training and evaluation
        logger.info("Training time-series model.")


        # scale data to the range of [0, 1]
        trainData = trainData.astype("float32") / 255.0
        testData = testData.astype("float32") / 255.0

        # wrap our model into a scikit-learn compatible classifier
        print("[INFO] initializing model...")

        # define a grid of the hyperparameter search space
        kernel_sizeLayerer_a = [16, 32, 64, 128, 256]
        kernel_sizeLayerer_b = [16, 32, 64, 128, 256]
        dropoutCovLayer_a= [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]
        # learnRate = [1e-2, 1e-3, 1e-4]
        add_one_layer= [False, True]
        # add_two_layer= [False, True]
        dropoutFullLayer_a = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]
        dropoutFullLayer_b = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]
        dropoutFullLayer_c = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]
        # batchSize = [4, 8, 16, 32]
        # epochs = [10, 20, 30, 40]


        # create a dictionary from the hyperparameter grid
        grid = dict(
        	kernel_sizeLayerer_a=kernel_sizeLayerer_a,
        	kernel_sizeLayerer_b=kernel_sizeLayerer_b,
        	dropoutCovLayer_a=dropoutCovLayer_a,
        	add_one_layer=add_one_layer,
        	dropoutFullLayer_a=dropoutFullLayer_a,
            dropoutFullLayer_b=dropoutFullLayer_b,
        	dropoutFullLayer_c=dropoutFullLayer_c
        )

        param_grid = ParameterGrid(grid)

        for params in param_grid:
            cur_model = get_mlp_model(**params)

            # print(model.summary())
            plot_model(
                cur_model, to_file="model_plot.png", show_shapes=True, show_layer_names=True
            )
            cur_trained_model = fit_model(
                work_dir,
                cur_model,
                data_type,
                ts_train_data,
                train_labs,
                train_s,
                ts_val_data,
                val_labs,
                val_s,
                class_weights,
                ua.experiment_name,
            )
            evaluate_model( ### need to modify to store fitted results
                cur_trained_model,
                ts_test_data,
                test_labs,
                test_s,
                work_dir,
                ua.experiment_name,
                data_type,
                lab_dict,
            )


        logger.info(f"SP Data shape (samples, haps): {sp_train_data.shape}")
        # Use only the final timepoint
        sp_train_data = np.squeeze(ts_train_data[:, -1, :])
        sp_val_data = np.squeeze(ts_val_data[:, -1, :])
        sp_test_data = np.squeeze(ts_test_data[:, -1, :])
        logger.info(f"SP Data shape (samples, haps): {sp_train_data.shape}")

        sp_datadim = sp_train_data.shape[-1]

        for params in param_grid:
            cur_model = model.__init__(self, **params)

            sp_datadim = sp_train_data.shape[-1]
            model = create_1Samp_model(sp_datadim, len(lab_dict))

            trained_model = fit_model(
                work_dir,
                model,
                data_type,
                sp_train_data,
                train_labs,
                train_s,
                sp_val_data,
                val_labs,
                val_s,
                class_weights,
                ua.experiment_name,
            )
            evaluate_model(
                trained_model,
                sp_test_data,
                test_labs,
                test_s,
                work_dir,
                ua.experiment_name,
                data_type,
                lab_dict,
            )
            )



# # initialize a random search with a 3-fold cross-validation and then
# # start the hyperparameter search process
# print("[INFO] performing grid search...")
# searcher = GridSearchCV(estimator=model, n_jobs=-1, cv=1
# 	param_distributions=grid, scoring="accuracy")
# searchResults = searcher.fit(train_data, [train_labs, train_s])
#
#
# # summarize grid search information
# bestScore = searchResults.best_score_
# bestParams = searchResults.best_params_
# print("[INFO] best score is {:.2f} using {}".format(bestScore,
# 	bestParams))
#
# # extract the best model, make predictions on our data, and show a
# # classification report
# print("[INFO] evaluating the best model...")
# bestModel = searchResults.best_estimator_
# accuracy = bestModel.score(testData, testLabels)
# print("accuracy: {:.2f}%".format(accuracy * 100))



import sys
import numpy as np
import argparse



def getRepAndGen(headerEntry):
    genInfo, repInfo = headerEntry.split("_")[-2:]
    rep = int(repInfo)
    if genInfo == "Base":
        gen = 0
    else:
        gen = int(genInfo.lstrip("F"))
    return rep, gen


def getFreqs(baseCounts):
    aCount, tCount, cCount, gCount, nCount, delCount = [
        int(x) for x in baseCounts.split(":")
    ]
    counts = [aCount, tCount, cCount, gCount, delCount]
    denom = sum(counts)

    return [x / denom for x in counts]


def getWinningAlleleIndex(allFreqsForSnp, mode="final"):
    assert allFreqsForSnp.shape == (7, 5)  # 7 timepoints X freqs for 5 possible alleles

    freqScores = []
    for alleleIndex in range(allFreqsForSnp.shape[1]):
        if mode == "final":
            freqScore = allFreqsForSnp[-1][alleleIndex]
        elif mode == "velocity":
            freqScore = allFreqsForSnp[-1][alleleIndex] - allFreqsForSnp[0][alleleIndex]
        freqScores.append(freqScore)

    return np.argmax(freqScores)


def getMostCommonOtherAlleleIndex(allFreqsForSnp, winningAlleleIndex):
    assert allFreqsForSnp.shape == (7, 5)  # 7 timepoints X freqs for 5 possible alleles

    totalFreqs = []
    for alleleIndex in range(allFreqsForSnp.shape[1]):
        if alleleIndex == winningAlleleIndex:
            totalFreqs.append(-1)
        else:
            totalFreq = sum(allFreqsForSnp[:, alleleIndex])
            totalFreqs.append(totalFreq)

    return np.argmax(totalFreqs)


def encodeFreqsInPlace(freqArray):
    for snpIndex in range(len(freqArray[0])):
        allFreqsForSnp = []

        for gen in range(len(freqArray)):
            freqs = freqArray[gen][snpIndex]
            allFreqsForSnp.append(freqs)

        allFreqsForSnp = np.array(allFreqsForSnp)
        winningAlleleIndex = getWinningAlleleIndex(allFreqsForSnp, mode="velocity")
        otherAlleleIndex = getMostCommonOtherAlleleIndex(
            allFreqsForSnp, winningAlleleIndex
        )

        for gen in range(len(freqArray)):
            freqs = freqArray[gen][snpIndex]
            winningFreq = freqs[winningAlleleIndex]
            otherFreq = freqs[otherAlleleIndex]

            freq = winningFreq / (winningFreq + otherFreq)
            freqArray[gen][snpIndex] = freq
        # assert winningFreq > 0 and winningFreq >= otherFreq

def get_header(headerfile):
    with open(headerfile, "r") as hfile:
        header = hfile.readline().strip().split()
        
    return header

def main(ua):
    agp = argparse.ArgumentParser()
    agp.add_argument("-i", "--infile", help="Sync file to convert to timesweeper NPZ format.")
    agp.add_argument("-c", "--chrom", help="Chromosome to pull out of sync file.")
    agp.add_argument("-o", "--outdir", help="Output directory.")
    agp.add_argument("-w", "--window_size", help="Size of windows to extract.")
    agp.add_argument("-h", "--headerfile", help="File containing a single line - the header of the syncfile. Must be provided if header is not present in the syncfile.")
    ua = agp.parse_args()
    
    inFileName = ua.infile
    outDir = ua.outdir 
    winSize = ua.window_size 
    targetChrom = ua.chrom 
    
    if ua.headerfile:
       header = get_header(ua.headerfile)
    else:
        header = get_header(inFileName) 
    
    freqs = {}
    for i in range(len(header)):
        if "Dsim" in header[i]:
            rep, gen = getRepAndGen(header[i])
            if not rep in freqs:
                freqs[rep] = {}
            if not gen in freqs[rep]:
                freqs[rep][gen] = {}


    sys.stderr.write("reading snps and freqs\n")
    positions = {}
    with open(inFileName, "rt") as inFile:
        for line in inFile:
            line = line.strip().split()
            chrom = line[0]
            pos = int(line[1])
            if chrom == targetChrom:
                if not chrom in positions:
                    positions[chrom] = []
                    for rep in freqs:
                        for gen in freqs[rep]:
                            freqs[rep][gen][chrom] = []

                for i in range(len(line)):
                    if ":" in line[i]:
                        rep, gen = getRepAndGen(header[i])
                        currFreqs = getFreqs(line[i])
                        freqs[rep][gen][chrom].append(currFreqs)
                positions[chrom].append(pos)
    sys.stderr.write("got all snps and freqs\n")


    sys.stderr.write("formatting output\n")
    for rep in freqs:
        for chrom in positions:
            outFileName = f"{outDir}/dsim_chrom_{chrom}_rep_{rep}.npz"

            assert len(positions[chrom]) == len(freqs[rep][gen][chrom])

            freqArray = []
            for gen in sorted(freqs[rep]):
                freqArray.append([])
                for posIndex in range(len(positions[chrom])):
                    currFreqs = freqs[rep][gen][chrom][posIndex]
                    freqArray[-1].append(currFreqs)

            encodeFreqsInPlace(freqArray)
            freqArray = np.array(freqArray)
            numGens = len(freqs[rep])
            assert freqArray.shape == (numGens, len(positions[chrom]))

            allFreqWins = []
            allPosWins = []
            startingIndices = range(len(positions[chrom]) - winSize)
            for i in startingIndices:
                currWin = freqArray[:, i : i + winSize]
                allFreqWins.append(currWin)
                currPositions = positions[chrom][i : i + winSize]
                allPosWins.append(currPositions)
            assert i + winSize == len(positions[chrom]) - 1

            allFreqWins = np.array(allFreqWins)
            allPosWins = np.array(allPosWins)
            assert allFreqWins.shape == (len(startingIndices), numGens, winSize)
            assert allPosWins.shape == (len(startingIndices), winSize)

            np.savez(outFileName, aftIn=allFreqWins, aftInPosition=allPosWins)
            
    sys.stderr.write("all done!\n")

    
if __name__=="__main__":
    main()


import logging
import multiprocessing as mp
import os
import subprocess
import sys
from glob import glob

import numpy as np
import yaml

from timesweeper import process_vcfs as pv

logging.basicConfig()
logger = logging.getLogger("sim_custom")


def read_config(yaml_file):
    """Reads in the YAML config file."""
    with open(yaml_file, "r") as infile:
        yamldata = yaml.safe_load(infile)

    return yamldata


# Simulation
def randomize_selCoeff_uni(lower_bound=0.00025, upper_bound=0.25):
    """Draws selection coefficient from log uniform dist to vary selection strength."""
    rng = np.random.default_rng(
        np.random.seed(int.from_bytes(os.urandom(4), byteorder="little"))
    )

    return rng.uniform(lower_bound, upper_bound, 1)[0]


def randomize_sampGens(num_timepoints, dev=50, span=200):
    rng = np.random.default_rng(
        np.random.seed(int.from_bytes(os.urandom(4), byteorder="little"))
    )
    start = round(rng.uniform(-dev, dev, 1)[0])
    if num_timepoints == 1:
        sampGens = [start + span]
    else:
        sampGens = [
            round(i) for i in np.linspace(start, start + span + 1, num_timepoints)
        ]

    return sampGens


def make_d_block(
    sweep,
    outFileVCF,
    outFileMS,
    dumpfile,
    num_sample_points,
    inds_per_tp,
    physLen,
    verbose=False,
):
    """
    This is meant to be a very customizeable block of text for adding custom args to SLiM as constants.
    Can add other functions to this module and call them here e.g. pulling selection coeff from a dist.
    This block MUST INCLUDE the 'sweep' and 'outFile' params, and at the very least the outFile must be used as output for outputVCFSample.
    Please note that when feeding strings as a constant you must escape them since this is a shell process.
    """
    selCoeff = randomize_selCoeff_uni()
    sampGens = [str(i) for i in randomize_sampGens(num_sample_points)]

    d_block = f"""\
    -d "sweep='{sweep}'" \
    -d "outFileVCF='{outFileVCF}'" \
    -d "outFileMS='{outFileMS}'" \
    -d "dumpFile='{dumpfile}'" \
    -d selCoeff={selCoeff} \
    -d sampGens='c({','.join(sampGens)})' \
    -d numSamples={num_sample_points} \
    -d sampleSizePerStep={inds_per_tp} \
    -d physLen={physLen} \
    -d seed={np.random.randint(0, 1e16)} \
    """
    if verbose:
        logger.info(f"Using the following constants with SLiM: {d_block}")

    return d_block


def simulate(slim_path, d_block, slimfile, logfile):
    cmd = " ".join(["time", slim_path, d_block, slimfile, ">>", logfile]).replace(
        "    ", ""
    )
    with open(logfile, "w") as ofile:
        ofile.write(cmd)

    try:
        subprocess.run(cmd, shell=True)
    except subprocess.CalledProcessError as e:
        logger.error(e.output)


def simulate_prep(
    vcf_file, num_sample_points, slimfile, slim_path, d_block, logfile, dumpFile
):
    simulate(slim_path, d_block, slimfile, logfile)
    os.remove(dumpFile)

    pv.process_vcfs(vcf_file, num_sample_points)
    os.remove(vcf_file)


def main(ua):
    """
    For simulating non-stdpopsim SLiMfiles.
    Currently only works with 1 pop models where m2 is the sweep mutation.
    Otherwise everything else is identical to stdpopsim version, just less complex.

    Generalized block of '-d' arguments to give to SLiM at the command line allow for
    flexible script writing within the context of this wrapper. If you write your SLiM script
    to require args set at runtime, this should be easily modifiable to do what you need and
    get consistent results to plug into the rest of the workflow.

    The two things you *will* need to specify in your '-d' args to SLiM (and somewhere in the slim script) are:
    - sweep [str] One of "neut", "sdn", or "ssv". If you're testing only a neut/sdn model,
        make the ssv a dummy switch for the neutral scenario.
    - outFile [path] You will need to define this as a population outputVCFSample input, with replace=F and append=T.
        This does *not* need to be specified by you in the custom -d block, it will be standardized to work with the rest of the pipeline using work_dir.
        example line for slim script: `p1.outputVCFSample(sampleSizePerStep, replace=F, append=T, filePath=outFile);`

    """
    yaml_data = read_config(ua.yaml_file)
    (
        work_dir,
        slim_file,
        slim_path,
        reps,
        rep_range,
        num_sample_points,
        inds_per_tp,
        physLen,
    ) = (
        yaml_data["work dir"],
        yaml_data["slimfile"],
        yaml_data["slim path"],
        yaml_data["reps"],
        ua.rep_range,
        yaml_data["num_sample_points"],
        yaml_data["inds_per_tp"],
        yaml_data["physLen"],
    )

    vcf_dir = f"{work_dir}/vcfs"
    ms_dir = f"{work_dir}/mss"
    dumpfile_dir = f"{work_dir}/dumpfiles"
    logfile_dir = f"{work_dir}/logs"

    sweeps = ["neut", "sdn", "ssv"]

    for i in [vcf_dir, dumpfile_dir, logfile_dir]:
        for sweep in sweeps:
            os.makedirs(f"{i}/{sweep}", exist_ok=True)

    mp_args = []
    # Inject info into SLiM script and then simulate, store params for reproducibility
    if rep_range:  # Take priority
        replist = range(int(rep_range[0]), int(rep_range[1]) + 1)
    else:
        replist = range(reps)

    for rep in replist:
        for sweep in sweeps:
            outFileVCF = f"{vcf_dir}/{sweep}/{rep}.multivcf"
            outFileMS = f"{ms_dir}/{sweep}/{rep}.multiMsOut"
            dumpFile = f"{dumpfile_dir}/{sweep}/{rep}.dump"
            logFile = f"{logfile_dir}/{sweep}/{rep}.log"

            d_block = make_d_block(
                sweep,
                outFileVCF,
                outFileMS,
                dumpFile,
                num_sample_points,
                inds_per_tp,
                physLen,
                False,
            )

            mp_args.append(
                (
                    outFileVCF,
                    num_sample_points,
                    slim_file,
                    slim_path,
                    d_block,
                    logFile,
                    dumpFile,
                )
            )

    pool = mp.Pool(processes=ua.threads)
    pool.starmap(simulate_prep, mp_args, chunksize=1)



from setuptools import setup

if __name__ == "__main__":
    with open("README.md", "r") as readme:
        readme_txt = readme.read()
    
    setup(long_description=readme_txt, long_description_content_type='text/markdown')



import multiprocessing as mp
import os
import subprocess
from glob import glob
from itertools import cycle


# VCF Processing
def read_multivcf(input_vcf):
    """Reads in file and returns as list of strings."""
    with open(input_vcf, "r") as input_file:
        raw_lines = [i.strip() for i in input_file.readlines()]

    return raw_lines


def split_multivcf(vcf_lines, header):
    """Splits the lines of multi-vcf file into list of vcf entries by <header> using itertools."""
    header_idxs = [i for i in range(len(vcf_lines)) if vcf_lines[i] == header]

    split_vcfs = []
    for idx in range(len(header_idxs[:-1])):
        split_vcfs.append(vcf_lines[header_idxs[idx] : header_idxs[idx + 1]])

    split_vcfs.append(vcf_lines[header_idxs[-1] :])

    return split_vcfs


def write_vcfs(vcf_lines, vcf_dir):
    """Writes list of vcf entries to numerically-sorted vcf files."""
    for idx, lines in enumerate(vcf_lines):
        with open(os.path.join(vcf_dir, f"{idx}.vcf"), "w") as outfile:
            outfile.writelines("\n".join(lines))


def index_vcf(vcf):
    """
    Indexes and sorts vcf file.
    Commands are run separately such that processes complete before the next one starts.
    """
    bgzip_cmd = f"bgzip -c {vcf} > {vcf}.gz"
    tabix_cmd = f"tabix -f -p vcf {vcf}.gz"
    bcftools_cmd = f"bcftools sort -Ov {vcf}.gz | bgzip -f > {vcf}.sorted.gz"
    tabix_2_cmd = f"tabix -f -p vcf {vcf}.sorted.gz"
    subprocess.run(
        bgzip_cmd, shell=True, stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL
    )
    subprocess.run(
        tabix_cmd.split(), stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL
    )
    subprocess.run(
        bcftools_cmd, shell=True, stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL
    )
    subprocess.run(
        tabix_2_cmd.split(), stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL
    )


def merge_vcfs(vcf_dir):
    num_files = len(glob(f"{vcf_dir}/*.vcf.sorted.gz"))
    if num_files == 1:
        cmd = f"""zcat {f"{vcf_dir}/0.vcf.sorted.gz"} > {vcf_dir}/merged.vcf"""

    else:
        cmd = f"""bcftools merge -Ov -0 \
                --force-samples --info-rules 'MT:join,S:join' \
                {" ".join([f"{vcf_dir}/{i}.vcf.sorted.gz" for i in range(num_files)])} > \
                {vcf_dir}/merged.vcf \
                """

    subprocess.run(cmd, shell=True)


def get_num_inds(vcf_file):
    num_ind = subprocess.check_output(
        """awk '{if ($1 == "#CHROM"){print NF-9; exit}}' """ + vcf_file, shell=True,
    )
    return int(num_ind)


def cleanup_intermed(vcf_dir):
    for ifile in glob(f"{vcf_dir}/*"):
        if "merged" not in ifile and "final" not in ifile:
            pass
            os.remove(ifile)


def make_vcf_dir(input_vcf):
    """Creates directory named after vcf basename."""
    dirname = os.path.basename(input_vcf).split(".")[0]
    dirpath = os.path.dirname(input_vcf)
    vcf_dir = os.path.join(dirpath, dirname)
    if os.path.exists(vcf_dir):
        for ifile in glob(f"{vcf_dir}/*"):
            os.remove(ifile)

    os.makedirs(vcf_dir, exist_ok=True)

    final_vcf = f"{input_vcf}.final"
    if os.path.exists(final_vcf):
        os.rename(final_vcf, f"{vcf_dir}/{final_vcf.split('/')[-1]}")

    return vcf_dir


def process_vcfs(input_vcf, num_tps):
    try:
        # Split into multiples after SLiM just concats to same file
        raw_lines = read_multivcf(input_vcf)
        split_lines = split_multivcf(raw_lines, "##fileformat=VCFv4.2")
        if len(split_lines) > 0:
            split_lines = split_lines[len(split_lines) - num_tps :]

            # Creates subdir for each rep
            vcf_dir = make_vcf_dir(input_vcf)
            write_vcfs(split_lines, vcf_dir)

            # Now index and merge
            [index_vcf(vcf) for vcf in glob(f"{vcf_dir}/*.vcf")]
            merge_vcfs(vcf_dir)

            cleanup_intermed(vcf_dir)

    except Exception as e:
        print(f"[ERROR] Couldn't process {e}")
        pass


def main(ua):
    vcflist = glob(f"{ua.in_dir}/*/*/*.multivcf")

    with mp.Pool(ua.threads) as p:
        p.map(vcflist, cycle([ua.num_tps]))

    print("[INFO] Done")
